import os
import pathlib
import shutil
import time
from typing import List, Union

from ada.config import get_logger

logger = get_logger()


class SIZE_UNIT:
    """Enum for size units"""

    BYTES = 1
    KB = 2
    MB = 3
    GB = 4


def copy_bulk(files, destination_dir, substitution_map=None):
    """
    Use shutil to copy a list of files to a specified destination directory. Can also parse in a substitution map (a
    dict with key: value substitution for specified files

    :param files:
    :param destination_dir:
    :param substitution_map:
    :return:
    """
    if os.path.isdir(destination_dir):
        shutil.rmtree(destination_dir)
        time.sleep(1)
    os.makedirs(destination_dir, exist_ok=True)

    for f in files:
        fname = os.path.basename(f)
        dest_file = os.path.join(destination_dir, fname)
        edited = False
        if substitution_map is not None:
            if fname in substitution_map.keys():
                edited = True
                with open(f, "r") as d:
                    in_str = d.read()
                in_str = in_str.replace(substitution_map[fname][0], substitution_map[fname][1])
                with open(dest_file, "w") as d:
                    d.write(in_str)
        if edited is False:
            shutil.copy(f, dest_file)


def convert_unit(size_in_bytes, unit):
    """Convert the size from bytes to other units like KB, MB or GB"""
    if unit == SIZE_UNIT.KB:
        return size_in_bytes / 1024
    elif unit == SIZE_UNIT.MB:
        return size_in_bytes / (1024**2)
    elif unit == SIZE_UNIT.GB:
        return size_in_bytes / (1024**3)
    else:
        return size_in_bytes


def get_dir_size(root_directory: pathlib.Path):
    return sum(f.stat().st_size for f in root_directory.glob("**/*") if f.is_file())


def get_file_size(file_name, size_type=SIZE_UNIT.MB):
    """Get file in size in given unit like KB, MB or GB"""
    size = os.path.getsize(file_name)
    return convert_unit(size, size_type)


def get_short_path_name(long_name):
    """Gets the short path name of a given long path (https://stackoverflow.com/a/23598461/200291)"""
    import ctypes
    from ctypes import wintypes

    _GetShortPathNameW = ctypes.windll.kernel32.GetShortPathNameW
    _GetShortPathNameW.argtypes = [wintypes.LPCWSTR, wintypes.LPWSTR, wintypes.DWORD]
    _GetShortPathNameW.restype = wintypes.DWORD

    output_buf_size = 0
    while True:
        output_buf = ctypes.create_unicode_buffer(output_buf_size)
        needed = _GetShortPathNameW(long_name, output_buf, output_buf_size)
        if output_buf_size >= needed:
            return output_buf.value
        else:
            output_buf_size = needed


def get_unc_path(path_) -> str:
    """Will try to convert path string to UNC path"""
    import win32wnet

    if path_[0].lower() == "c":
        return path_
    else:
        try:
            out_path = win32wnet.WNetGetUniversalName(path_)
            return out_path
        except BaseException as e:
            logger.error(e)
            return path_


def get_list_of_files(
    dir_path,
    file_ext=None,
    strict=False,
    filter_path_contains: Union[None, List[str], str] = None,
    keep_path_contains: Union[None, List[str], str] = None,
) -> list[str]:
    """Get a list of files and sub directories for a given directory"""
    all_files = []
    list_of_file = sorted(os.listdir(dir_path), key=str.lower)

    # Iterate over all the entries
    for entry in list_of_file:
        # Create full path
        full_path = os.path.join(dir_path, entry).replace(os.sep, "/")
        # If entry is a directory then get the list of files in this directory
        if os.path.isdir(full_path):
            all_files += get_list_of_files(full_path, file_ext, strict, filter_path_contains)
        else:
            if filter_path_contains is not None:
                if type(filter_path_contains) is str:
                    filter_path_contains = [filter_path_contains]
                skip_it = False
                for f in filter_path_contains:
                    if f in full_path:
                        skip_it = True
                        break
                if skip_it:
                    continue
            if keep_path_contains is not None:
                if type(keep_path_contains) is str:
                    keep_path_contains = [keep_path_contains]
                skip_it = True
                for f in keep_path_contains:
                    if f in full_path:
                        skip_it = False
                        break
                if skip_it:
                    continue
            all_files.append(full_path)

    if file_ext is not None:
        all_files = [f for f in all_files if f.endswith(file_ext)]

    if len(all_files) == 0:
        msg = f'Files with "{file_ext}"-extension is not found in "{dir_path}" or any sub-folder.'
        if strict:
            raise FileNotFoundError(msg)
        else:
            logger.info(msg)

    return all_files


def getfileprop(filepath: str) -> dict:
    """Read all properties of a local file and return them as a dictionary"""
    import win32api

    filepath = str(filepath)
    propNames = (
        "Comments",
        "InternalName",
        "ProductName",
        "CompanyName",
        "LegalCopyright",
        "ProductVersion",
        "FileDescription",
        "LegalTrademarks",
        "PrivateBuild",
        "FileVersion",
        "OriginalFilename",
        "SpecialBuild",
    )

    props = {"FixedFileInfo": None, "StringFileInfo": None, "FileVersion": None}

    try:
        # backslash as parm returns dictionary of numeric info corresponding to VS_FIXEDFILEINFO struc
        fixedInfo = win32api.GetFileVersionInfo(filepath, "\\")
        props["FixedFileInfo"] = fixedInfo
        props["FileVersion"] = "%d.%d.%d.%d" % (
            fixedInfo["FileVersionMS"] / 65536,
            fixedInfo["FileVersionMS"] % 65536,
            fixedInfo["FileVersionLS"] / 65536,
            fixedInfo["FileVersionLS"] % 65536,
        )

        # \VarFileInfo\Translation returns list of available (language, codepage)
        # pairs that can be used to retreive string info. We are using only the first pair.
        lang, codepage = win32api.GetFileVersionInfo(filepath, "\\VarFileInfo\\Translation")[0]

        # any other must be of the form \StringfileInfo\%04X%04X\parm_name, middle
        # two are language/codepage pair returned from above

        strInfo = {}
        for propName in propNames:
            strInfoPath = "\\StringFileInfo\\%04X%04X\\%s" % (lang, codepage, propName)
            strInfo[propName] = win32api.GetFileVersionInfo(filepath, strInfoPath)

        props["StringFileInfo"] = strInfo
    except Exception as e:
        logger.error(f'Unable to Read file properties due to "{e}"')
        pass

    return props


def path_leaf(path):
    import ntpath

    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

from ada.sat.readers.bsplinesurface import create_bsplinesurface_from_sat
from ada.sat.readers.face import PlateFactory

if TYPE_CHECKING:
    pass


class SatReader:
    def __init__(self, sat_file):
        self.f = open(sat_file, "r")
        self.lineno = 0
        self.header = ""

    def _read_line(self):
        try:
            self.lineno += 1
            return next(self.f)
        except StopIteration:
            self.f.close()
            raise StopIteration

    def __next__(self):
        line = self._read_line()

        if self.lineno == 1:
            header = line
            while self.lineno <= 2:
                header += self._read_line()
            return header

        if "#" in line:
            return line

        while True:
            line += self._read_line()
            if "#" in line:
                break

        return line

    def __iter__(self):
        return self


class SatStore:
    def __init__(self):
        self.sat_data = dict()

    def add(self, sat_id: int | str, sat_object_data: str):
        if isinstance(sat_id, str):
            sat_id = float(int(sat_id))
        self.sat_data[sat_id] = sat_object_data

    def get(self, sat_id: int | str) -> list[str]:
        if isinstance(sat_id, str):
            if sat_id.startswith("$"):
                sat_id = sat_id.replace("$", "")
            sat_id = float(int(sat_id))
        return self.sat_data[sat_id].split()

    def get_name(self, sat_id: int | str) -> str:
        res = self.get(sat_id)
        ref_type = res[1]
        if ref_type.startswith("string"):
            return res[-2]
        elif ref_type.startswith("position"):
            return self.get_name(res[4])
        else:
            raise NotImplementedError(f"Unknown reference type: {ref_type}")

    def iter(self):
        for sat_id in sorted(self.sat_data.keys()):
            yield self.sat_data[sat_id]


class SatReaderFactory:
    def __init__(self, sat_file):
        self.sat_file = sat_file
        self.entities = dict()
        self.sat_store = SatStore()
        self.plate_factory = PlateFactory(self.sat_store)
        self.header = ""

    def interpret_sat_object_data(self, sat_object_data: str):
        geom_id, sat_type = sat_object_data.split()[0:2]
        geom_id = geom_id.replace("-", "")

        if sat_type == "spline-surface":
            self.entities[geom_id] = create_bsplinesurface_from_sat(sat_object_data)
        elif sat_type == "face":
            self.entities[geom_id] = self.plate_factory.get_face_name_and_points(sat_object_data)
        else:
            self.entities[geom_id] = sat_object_data

    def store_sat_object_data(self):
        sat_reader = SatReader(self.sat_file)
        self.header = next(sat_reader)
        for sat_object_str in sat_reader:
            geom_id, sat_type = sat_object_str.split()[0:2]
            geom_id = geom_id.replace("-", "")
            self.sat_store.add(geom_id, sat_object_str)

    def iter_faces(self):
        if len(self.sat_store.sat_data) == 0:
            self.store_sat_object_data()

        for sat_object_data in self.sat_store.iter():
            geom_id, sat_type = sat_object_data.split()[0:2]
            if "face" == sat_type:
                yield sat_object_data

    def iter_flat_plates(self) -> Iterable[tuple[str, list[tuple[float, float, float]]]]:
        for face_data in self.iter_faces():
            pl = self.plate_factory.get_face_name_and_points(face_data)
            if pl is None:
                continue
            yield pl

    def read_data(self):
        self.store_sat_object_data()
        for sat_object in self.sat_store.iter():
            self.interpret_sat_object_data(sat_object)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from ada.base.changes import ChangeAction
from ada.config import get_logger
from ada.ifc.utils import create_guid

if TYPE_CHECKING:
    from ada.base.physical_objects import BackendGeom

logger = get_logger()


@dataclass
class PresentationLayer:
    name: str
    description: str
    members: list[BackendGeom] = field(default_factory=list)
    change_type: ChangeAction = ChangeAction.NOTDEFINED
    identifier: str = field(default_factory=create_guid)


@dataclass
class PresentationLayers:
    layers: dict[str, PresentationLayer] = field(default_factory=dict)

    def add_layer(self, layer: str | PresentationLayer, description: str = None) -> PresentationLayer:
        if isinstance(layer, PresentationLayer):
            existing_layer = self.layers.get(layer.name)
        else:
            existing_layer = self.layers.get(layer)

        if existing_layer is not None:
            raise ValueError(f'Existing Layer with name="{layer}": {existing_layer}')

        if isinstance(layer, PresentationLayer):
            new_layer = layer
        else:
            new_layer = PresentationLayer(layer, description=description, change_type=ChangeAction.ADDED)

        self.layers[layer] = new_layer

        return new_layer

    def get_by_name(self, layer_name) -> None | PresentationLayer:
        return self.layers.get(layer_name)

    def add_object(self, obj, layer: str):
        from ada import Part

        layer_obj = self.get_by_name(layer)
        if layer_obj is None:
            logger.info(f'Layer "{layer}" does not exist. So creating a new layer')
            layer_obj = self.add_layer(layer)
        else:
            layer_obj.change_type = ChangeAction.MODIFIED

        if isinstance(obj, Part):
            for geom in obj.get_all_physical_objects():
                layer_obj.members.append(geom)
        else:
            layer_obj.members.append(obj)

    def remove_layer_and_delete_objects(self, layer: str):
        layer_obj = self.layers.get(layer)
        for mem in layer_obj.members:
            mem.change_type = ChangeAction.DELETED

        self.layers.pop(layer)



import os
from typing import TYPE_CHECKING

from ada.fem.conversion_utils import convert_ecc_to_mpc, convert_hinges_2_couplings

from .write_constraints import constraints_str
from .write_elements import elements_str
from .write_masses import masses_str
from .write_nodes import nodes_str, rp_str
from .write_sections import sections_str
from .write_sets import elsets_str, nsets_str
from .write_springs import springs_str
from .write_surfaces import surfaces_str

if TYPE_CHECKING:
    from ada import Assembly, Part


def write_all_parts(assembly: "Assembly", analysis_dir):
    for part in assembly.get_all_subparts():
        if len(part.fem.elements) == 0:
            continue

        if assembly.convert_options.hinges_to_coupling is True:
            convert_hinges_2_couplings(part.fem)

        if assembly.convert_options.ecc_to_mpc is True:
            convert_ecc_to_mpc(part.fem)

        write_part_bulk(part, analysis_dir)


def write_part_bulk(part_in: "Part", analysis_dir):
    bulk_path = analysis_dir / f"bulk_{part_in.name}"
    bulk_file = bulk_path / "aba_bulk.inp"
    os.makedirs(bulk_path, exist_ok=True)

    if part_in.fem.initial_state is not None:
        with open(bulk_file, "w") as d:
            d.write("** This part is replaced by an initial state step")
        return None

    with open(bulk_file, "w") as d:
        d.write(write_abaqus_part_str(part_in))


def write_abaqus_part_str(part: "Part") -> str:
    fem = part.fem
    return f"""** Abaqus Part {part.name}
** Exported using ADA OpenSim
{nodes_str(fem)}
{elements_str(fem, False)}
{rp_str(fem)}
{elsets_str(fem, False)}
{nsets_str(fem, False)}
{sections_str(fem)}
{masses_str(fem, False)}
{surfaces_str(fem, False)}
{constraints_str(fem, False)}
{springs_str(fem)}""".rstrip()


def instance_move_str(self):
    if self.part.fem.metadata["move"] is not None:
        move = self.part.fem.metadata["move"]
        mo_str = "\n " + ", ".join([str(x) for x in move])
    else:
        mo_str = "\n 0.,        0.,           0."

    if self.part.fem.metadata["rotate"] is not None:
        rotate = self.part.fem.metadata["rotate"]
        vecs = ", ".join([str(x) for x in rotate[0]])
        vece = ", ".join([str(x) for x in rotate[1]])
        angle = rotate[2]
        move_str = """{move_str}\n {vecs}, {vece}, {angle}""".format(move_str=mo_str, vecs=vecs, vece=vece, angle=angle)
    else:
        move_str = "" if mo_str == "0.,        0.,           0." else mo_str
    return move_str



import xml.etree.ElementTree as ET
from dataclasses import dataclass
from typing import Dict, Iterable, List, Union

import numpy as np

from ada import Beam, Part, Section
from ada.concepts.connections import JointBase
from ada.concepts.containers import Connections
from ada.core.vector_utils import vector_length


@dataclass
class GenieJoint(JointBase):
    def __init__(self, name, chords, braces, parent: Connections):
        self.chords = chords
        self.braces = braces

        members = chords + braces
        self.eval_members(members)
        center = self.centre
        super(GenieJoint, self).__init__(name, members, center, parent)

    def _init_check(self, members):
        return None

    def contains_stiffener(self):
        chord_sec_types = [bm.section.type for bm in self.beams]
        return True if Section.TYPES.ANGULAR in chord_sec_types else False

    def eval_members(self, members) -> None:
        node_map = dict()
        len_map = dict()
        center = np.array([0, 0, 0], dtype=float)
        for mem in members:
            n1 = mem.n1
            n2 = mem.n2
            center += n1.p
            center += n2.p
            vlen = vector_length(n2.p - n1.p)
            if vlen > 10:
                continue
            len_map[mem] = vlen
            if n1 not in node_map.keys():
                node_map[n1] = []
            if n2 not in node_map.keys():
                node_map[n2] = []
            node_map[n1].append(mem)
            node_map[n2].append(mem)
        center_val = center / (2 * len(members))
        nodes = [(n, items) for n, items in node_map.items() if len(items) > 2]
        if len(nodes) == 0 or len(nodes) > 1:
            # TODO: Should consider multiple segments of same beam and find closest segment to center_val
            raise NotImplementedError(f'Joint @"{center_val}" error. Multiple or zero central nodes found')
        self._centre, self._stubs = nodes[0]

    @property
    def stubs(self) -> List[Beam]:
        return self._stubs

    @property
    def chord_stubs(self) -> List[Beam]:
        return list(filter(lambda x: x in self.stubs, self.chords))

    @property
    def brace_stubs(self) -> List[Beam]:
        return list(filter(lambda x: x in self.stubs, self.braces))

    @property
    def stub_weight(self) -> float:
        weight_sum = 0
        for bm in self.stubs:
            weight_sum += bm.length * bm.section.properties.Ax * bm.material.model.rho
        return weight_sum


def get_bm(el: ET.Element, part):
    ref = el.attrib["beam_ref"]
    if ref not in part.beams.dmap.keys():
        return None
    i = 2
    potential_beams = [part.beams.from_name(ref)]
    suffix = "_E{}"
    eval_name = ref + suffix.format(i)
    while eval_name in part.beams.dmap.keys():
        potential_beams.append(part.beams.from_name(eval_name))
        i += 1
        eval_name = ref + suffix.format(i)

    return potential_beams


def get_joint(type_tag: ET.Element, con: Connections, part: Part) -> Union[None, GenieJoint]:
    from itertools import chain

    def not_none(x):
        return x is not None

    name = type_tag.attrib["name"]
    chords = list(filter(not_none, chain.from_iterable(get_bm(chord, part) for chord in type_tag.findall("chords/"))))
    braces = list(filter(not_none, chain.from_iterable(get_bm(elem, part) for elem in type_tag.findall("braces/"))))
    if str in [type(x) for x in braces] or str in [type(x) for x in chords]:
        raise ValueError()
    if len(chords) == 0 and len(braces) == 0:
        return None
    return GenieJoint(name, chords, braces, con)


def get_joints(xml_root, part: Part) -> Connections:
    con = Connections(parent=part)
    joints = (get_joint(type_tag, con, part) for type_tag in xml_root.findall(".//frame_joint"))
    con.connections = filter_joints(joints)
    return con


def filter_joints(connections: Iterable[Union[None, GenieJoint]]) -> List[GenieJoint]:
    centers: Dict[tuple, GenieJoint] = dict()
    new_connections: List[GenieJoint] = []
    for con in connections:
        if con is None:
            continue
        con: GenieJoint
        c = tuple(con.centre)
        if c in centers.keys():
            raise NotImplementedError("How to merge coincident joints is not yet implemented")
        centers[c] = con
        new_connections.append(con)
    return new_connections



import datetime
import os


def get_now():
    d = datetime.datetime.now(datetime.timezone.utc).astimezone()
    date_in_str = "%a, %d %b %Y %H:%M:%S %Z %z"
    now_str = d.strftime(date_in_str)
    return now_str


def interpret_now(date_string):
    return datetime.datetime.strptime(date_string, "%a, %d %b %Y %H:%M:%S %Z %z")


def create_date_str(include_minutes=False):
    """
    Function for building a formatted timestamp string

    :param include_minutes: Extends the timing format to include minutes
    :return: formatted string <year>_<month>_<day> + optionally (_<hour>_<minute>)
    """
    du = datetime.datetime.now()
    dyear = str(du.year)
    dyear2 = "" + dyear[2] + "" + dyear[3]
    dmonth = "{:02d}".format(du.month)
    dday = "{:02d}".format(du.day)
    day_format = "" + str(dyear2) + "" + str(dmonth) + "" + str(dday)
    if include_minutes:
        day_format += "_" + str(du.hour) + ":" + str(du.minute)
    return day_format


def get_file_time_local(file_path):
    from datetime import datetime, timezone

    utc_time = datetime.fromtimestamp(os.path.getmtime(file_path), timezone.utc)
    return utc_time.astimezone()


def get_time_stamp_now():
    import pytz

    utc = pytz.UTC
    return utc.localize(datetime.datetime.utcnow())


def datetime_to_str(obj):
    return obj.isoformat()


def get_last_file_modified(file_dir, file_ext):
    from .file_system import get_list_of_files

    last_date = None
    for f in get_list_of_files(file_dir, file_ext):
        curr_date = get_file_time_local(f)
        if last_date is None:
            last_date = curr_date
        elif curr_date > last_date:
            last_date = curr_date
    return last_date


def datetime_from_str(obj_str):
    import dateutil.parser

    return dateutil.parser.parse(obj_str)



from __future__ import annotations

from dataclasses import dataclass, field
from itertools import chain
from typing import TYPE_CHECKING, Dict, Iterable, List, Tuple, Union

from ada.concepts.containers import Nodes
from ada.config import get_logger

from .containers import FemElements, FemSections, FemSets
from .sets import FemSet

if TYPE_CHECKING:
    from ada import Part
    from ada.concepts.points import Node
    from ada.concepts.stru_beams import Beam
    from ada.fem import (
        Amplitude,
        Bc,
        Connector,
        ConnectorSection,
        Constraint,
        Csys,
        Elem,
        FemSection,
        Interaction,
        InteractionProperty,
        Load,
        Mass,
        PredefinedField,
        Spring,
        StepEigen,
        StepExplicit,
        StepImplicit,
        StepSteadyState,
        Surface,
    )
    from ada.fem.results.common import Mesh
    from ada.fem.steps import Step

_step_types = Union["StepSteadyState", "StepEigen", "StepImplicit", "StepExplicit"]
logger = get_logger()


@dataclass
class InterfaceNode:
    node: Node
    constraint: Constraint = field(default=None)
    connector: Connector = field(default=None)


@dataclass
class FEM:
    name: str
    metadata: Dict = field(default_factory=dict)
    parent: Part = field(init=True, default=None)

    masses: Dict[str, Mass] = field(init=False, default_factory=dict)
    surfaces: Dict[str, Surface] = field(init=False, default_factory=dict)
    amplitudes: Dict[str, Amplitude] = field(init=False, default_factory=dict)
    connector_sections: Dict[str, ConnectorSection] = field(init=False, default_factory=dict)
    springs: Dict[str, Spring] = field(init=False, default_factory=dict)
    intprops: Dict[str, InteractionProperty] = field(init=False, default_factory=dict)
    interactions: Dict[str, Interaction] = field(init=False, default_factory=dict)
    predefined_fields: Dict[str, PredefinedField] = field(init=False, default_factory=dict)
    lcsys: Dict[str, Csys] = field(init=False, default_factory=dict)
    constraints: Dict[str, Constraint] = field(init=False, default_factory=dict)

    bcs: List[Bc] = field(init=False, default_factory=list)
    steps: List[Union[StepSteadyState, StepEigen, StepImplicit, StepExplicit]] = field(init=False, default_factory=list)

    nodes: Nodes = field(default_factory=Nodes, init=True)
    ref_points: Nodes = field(default_factory=Nodes, init=True)
    ref_sets: FemSets = field(default_factory=FemSets, init=True)

    elements: FemElements = field(default_factory=FemElements, init=True)
    sets: FemSets = field(default_factory=FemSets, init=True)
    sections: FemSections = field(default_factory=FemSections, init=True)
    initial_state: PredefinedField = field(default=None, init=True)
    subroutine: str = field(default=None, init=True)

    interface_nodes: List[Union[Node, InterfaceNode]] = field(init=False, default_factory=list)

    def __post_init__(self):
        self.nodes.parent = self
        self.elements.parent = self
        self.sets.parent = self
        self.sections.parent = self
        from ada.fem.options import FemOptions

        self._options = FemOptions()

    def add_elem(self, elem: Elem) -> Elem:
        elem.parent = self
        self.elements.add(elem)
        return elem

    def add_section(self, section: FemSection) -> FemSection:
        section.parent = self
        if section.elset.parent is None:
            if section.elset.name in self.elsets.keys():
                fs = self.elsets[section.elset.name]
            else:
                fs = self.sets.add(section.elset)
            if fs != section.elset:
                logger.info(f'Element set "{section.elset}" is replaced by {fs}')
                section.elset = fs
        if section.material.parent is None and self.parent is not None:
            self.parent.add_material(section.material)
        self.sections.add(section)
        return section

    def add_bc(self, bc: Bc) -> Bc:
        if bc.name in [b.name for b in self.bcs]:
            raise ValueError(f'BC with name "{bc.name}" already exists')

        bc.parent = self
        if bc.fem_set.parent is None:
            logger.debug("Bc FemSet has no parent. Adding to self")
            self.sets.add(bc.fem_set)

        self.bcs.append(bc)
        return bc

    def add_mass(self, mass: Mass) -> Tuple[Mass, FemSet]:
        mass.parent = self
        self.elements.add(mass)
        elset = self.sets.add(FemSet(mass.name + "_set", [mass], "elset"))
        mass.elset = elset
        return mass, elset

    def add_set(
        self,
        fem_set: FemSet,
        p=None,
        vol_box=None,
        vol_cyl=None,
        single_member=False,
        tol=1e-4,
    ) -> FemSet:
        """
        :param fem_set: A fem set object
        :param p: Single point (x,y,z)
        :param vol_box: Search by a box volume. Where p is (xmin, ymin, zmin) and vol_box is (xmax, ymax, zmax)
        :param vol_cyl: Search by cylindrical volume. Used together with p to find
                        nodes within cylinder inputted by [radius, height, thickness]
        :param single_member: Set True if you wish to keep only a single member
        :param tol: Point Tolerances. Default is 1e-4
        """
        fem_set.parent = self

        def append_members(nodelist):
            if single_member is True:
                fem_set.add_members([nodelist[0]])
            else:
                fem_set.add_members(nodelist)

        if fem_set.type != fem_set.TYPES.NSET or all(x is None for x in [p, vol_box, vol_cyl]):
            self.sets.add(fem_set)
            return fem_set

        nodes = self.nodes.get_by_volume(p, vol_box, vol_cyl, tol)
        if len(nodes) > 0:
            append_members(nodes)
            self.sets.add(fem_set)
            return fem_set

        if len(nodes) == 0 and self.parent is not None:
            assembly = self.parent.get_assembly()
            list_of_ps = assembly.get_all_subparts() + [assembly]
            for part in list_of_ps:
                nodes = part.fem.nodes.get_by_volume(p, vol_box, vol_cyl, tol)
                if len(nodes) == 0:
                    continue
                fem_set.parent = part.fem
                append_members(nodes)
                part.fem.add_set(fem_set)
                return fem_set

        raise Exception(f'No nodes found for femset "{fem_set.name}"')

    def add_step(self, step: _step_types) -> _step_types:
        """Add an analysis step to the assembly"""
        from ada.fem.steps import Step

        if len(self.steps) > 0:
            if self.steps[-1].type != Step.TYPES.EIGEN and step.type == Step.TYPES.COMPLEX_EIG:
                raise Exception("Complex eigenfrequency analysis step needs to follow eigenfrequency step.")
        step.parent = self
        for bc in step.bcs.values():
            if bc.amplitude is not None:
                if bc.amplitude.parent is None:
                    self.add_amplitude(bc.amplitude)
        self.steps.append(step)

        return step

    def add_interaction_property(self, int_prop: InteractionProperty) -> InteractionProperty:
        int_prop.parent = self
        self.intprops[int_prop.name] = int_prop
        return int_prop

    def add_interaction(self, interaction: Interaction) -> Interaction:
        interaction.parent = self
        self.interactions[interaction.name] = interaction
        if interaction.interaction_property.parent is None:
            self.add_interaction_property(interaction.interaction_property)
        return interaction

    def add_constraint(self, constraint: Constraint) -> Constraint:
        constraint.parent = self
        if constraint.m_set.parent is None:
            self.add_set(constraint.m_set)

        if constraint.s_set.parent is None:
            self.add_set(constraint.s_set)

        self.constraints[constraint.name] = constraint
        return constraint

    def add_lcsys(self, lcsys: Csys) -> Csys:
        if lcsys.name in self.lcsys.keys():
            raise ValueError("Local Coordinate system cannot have duplicate name")
        lcsys.parent = self
        self.lcsys[lcsys.name] = lcsys
        return lcsys

    def add_connector_section(self, connector_section: ConnectorSection) -> ConnectorSection:
        connector_section.parent = self
        self.connector_sections[connector_section.name] = connector_section
        return connector_section

    def add_connector(self, connector: Connector) -> Connector:
        connector.parent = self
        self.elements.add(connector)
        connector.csys.parent = self
        if connector.con_sec.parent is None:
            self.add_connector_section(connector.con_sec)
        self.add_set(FemSet(name=connector.name, members=[connector], set_type="elset"))
        return connector

    def add_rp(self, name: str, node: Node):
        """Adds a reference point in assembly with a specific name"""
        node.parent = self
        node_ = self.ref_points.add(node)
        fem_set = self.ref_sets.add(FemSet(name, [node_], "nset", parent=self))
        fem_set.metadata["internal"] = True
        return node_, fem_set

    def add_surface(self, surface: Surface) -> Surface:
        surface.parent = self
        self.surfaces[surface.name] = surface
        return surface

    def add_amplitude(self, amplitude: Amplitude) -> Amplitude:
        amplitude.parent = self
        self.amplitudes[amplitude.name] = amplitude
        return amplitude

    def add_predefined_field(self, pre_field: PredefinedField) -> PredefinedField:
        pre_field.parent = self
        self.predefined_fields[pre_field.name] = pre_field
        return pre_field

    def add_spring(self, spring: Spring) -> Spring:
        if spring.fem_set.parent is None:
            self.sets.add(spring.fem_set)
        self.springs[spring.name] = spring
        return spring

    def add_interface_nodes(self, interface_nodes: List[Union[Node, InterfaceNode]]):
        """Nodes used for interfacing between other parts. Pass a custom Constraint if specific coupling is needed"""
        from ada import Node

        for n in interface_nodes:
            n_in = InterfaceNode(n) if isinstance(n, Node) else n
            self.interface_nodes.append(n_in)

    def create_fem_elem_from_obj(self, obj, el_type=None) -> Elem:
        """Converts structural object to FEM elements. Currently only BEAM is supported"""
        from ada.fem.shapes import ElemType

        if type(obj) is not Beam:
            raise NotImplementedError(f'Object type "{type(obj)}" is not yet supported')

        el_type = ElemType.LINE if el_type is None else el_type

        res = self.nodes.add(obj.n1)
        if res is not None:
            obj.n1 = res
        res = self.nodes.add(obj.n2)
        if res is not None:
            obj.n2 = res

        elem = self.add_elem(Elem(None, [obj.n1, obj.n2], el_type))
        femset = self.add_set(FemSet(f"{obj.name}_set", [elem], FemSet.TYPES.ELSET))
        self.add_section(
            FemSection(
                f"d{obj.name}_sec",
                ElemType.LINE,
                femset,
                obj.material,
                obj.section,
                obj.ori[1],
            )
        )
        return elem

    def is_empty(self) -> bool:
        containers = [
            len(self.nodes),
            len(self.elements),
        ]

        for n_cont in containers:
            if n_cont != 0:
                return False

        return True

    def get_all_steps(self) -> list[Step]:
        assembly = self.parent.get_assembly()
        steps = []
        for p in assembly.get_all_parts_in_assembly(include_self=True):
            if len(p.fem.steps) == 0:
                continue
            steps += p.fem.steps
        return steps

    def get_all_bcs(self) -> Iterable[Bc]:
        """Get all the boundary conditions in the entire assembly"""
        assembly = self.parent.get_assembly()
        return chain.from_iterable(
            (
                assembly.fem.bcs,
                [bc for p in assembly.get_all_parts_in_assembly() for bc in p.fem.bcs],
            )
        )

    def get_all_masses(self) -> Iterable[Mass]:
        """Get all the Masses in the entire assembly"""
        assembly = self.parent.get_assembly()
        return chain.from_iterable(
            (
                assembly.fem.masses.values(),
                [mass for p in assembly.get_all_parts_in_assembly() for mass in p.fem.masses.values()],
            )
        )

    def get_all_loads(self) -> list[Load]:
        loads = []
        for step in self.steps:
            for load in step.loads:
                loads.append(load)
        return loads

    def to_mesh(self) -> Mesh:
        from ada.fem.results.common import Mesh

        fem_nodes = self.nodes.to_fem_nodes()
        elem_blocks = self.elements.to_elem_blocks()
        return Mesh(elem_blocks, fem_nodes)

    @property
    def instance_name(self):
        return self.name if self.name is not None else f"{self.parent.name}-1"

    @property
    def nsets(self):
        return self.sets.nodes

    @property
    def elsets(self):
        return self.sets.elements

    @property
    def options(self):
        return self._options

    def __add__(self, other: FEM):
        # Nodes
        nodid_max = self.nodes.max_nid if len(self.nodes) > 0 else 0
        if nodid_max > other.nodes.min_nid:
            other.nodes.renumber(int(nodid_max + 10))

        self.nodes.parent = self
        self.nodes += other.nodes

        # Elements
        elid_max = self.elements.max_el_id if len(self.elements) > 0 else 0

        if elid_max > other.elements.min_el_id:
            other.elements.renumber(int(elid_max + 10))

        self.elements += other.elements
        self.sections += other.sections
        self.sets += other.sets

        for bc in other.bcs:
            bc.parent = self
            self.bcs.append(bc)

        for con in other.constraints.values():
            con.parent = self
            self.constraints[con.name] = con

        for name, csys in other.lcsys.items():
            csys.parent = self
            self.lcsys[name] = csys

        for name, con_sec in other.connector_sections.items():
            con_sec.parent = self
            self.connector_sections[name] = con_sec

        for name, mass in other.masses.items():
            mass.parent = self
            self.masses[name] = mass

        for name, surface in other.surfaces.items():
            surface.parent = self
            self.surfaces[name] = surface

        if self.parent is None or other.parent is None:
            return self

        self.parent.materials += other.parent.materials

        return self

    def __repr__(self):
        return f"FEM({self.name}, Elements: {len(self.elements)}, Nodes: {len(self.nodes)})"



from __future__ import annotations

import pathlib
from subprocess import CompletedProcess

from ada.fem.formats.general import fem_executables
from ada.fem.formats.utils import default_fem_inp_path


def execute_fem(
    name, fem_format, scratch_dir, cpus, gpus, run_ext, metadata, execute, exit_on_complete, run_in_shell
) -> CompletedProcess | None:
    fem_inp_files = default_fem_inp_path(name, scratch_dir)
    exe_func = fem_executables.get(fem_format, None)
    inp_path = fem_inp_files.get(fem_format, None)

    if exe_func is None:
        if execute is False:
            return None
        else:
            raise NotImplementedError(f'The FEM format "{fem_format}" has no execute function')

    if inp_path is None:
        raise ValueError(f"FEM format '{fem_format}' is not supported")

    if isinstance(inp_path, str):
        inp_path = pathlib.Path(inp_path)

    out = exe_func(
        inp_path=inp_path,
        cpus=cpus,
        gpus=gpus,
        run_ext=run_ext,
        metadata=metadata,
        execute=execute,
        exit_on_complete=exit_on_complete,
        run_in_shell=run_in_shell,
    )

    if out is None:
        return None

    out_str = ""
    for out_stream in [out.stdout, out.stderr]:
        if out_stream is not None:
            out_str += out_stream

    with open(inp_path.parent / "run_log.txt", "w", encoding="utf8") as f:
        f.write(out_str)

    return out



# coding=utf-8
from __future__ import annotations

import os
import pathlib
import shutil
import zipfile
from decimal import ROUND_HALF_EVEN, Decimal
from typing import TYPE_CHECKING, Any, Dict, Union

import numpy as np

from ada.config import Settings, get_logger

if TYPE_CHECKING:
    from ada import Node

logger = get_logger()


class NewLine:
    def __init__(self, n, prefix=None, suffix=None):
        self.i = 0
        self.n = n
        self.prefix = prefix
        self.suffix = suffix

    def __iter__(self):
        return self

    def __next__(self):
        if self.i < self.n:
            self.i += 1
            return ""
        else:
            self.i = 0
            prefix = self.prefix if self.prefix is not None else ""
            suffix = self.suffix if self.suffix is not None else ""
            return prefix + "\n" + suffix


class Counter:
    def __init__(self, start: int = 1, prefix: str = None):
        self.i = start - 1
        self._prefix = prefix

    def set_i(self, i):
        self.i = i

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, value):
        self._prefix = value

    def __iter__(self):
        return self

    def __next__(self):
        self.i += 1
        return self.i if self._prefix is None else f"{self._prefix}{self.i}"


def random_color():
    from random import randint

    from OCC.Display.WebGl.jupyter_renderer import format_color

    return format_color(randint(0, 255), randint(0, 255), randint(0, 255))


def d2npy(node: Node) -> np.ndarray:
    """This method takes in a node object and returns a np.array."""
    return np.array([node.x, node.y, node.z], dtype=np.float)


def roundoff(x: float, precision=Settings.precision) -> float:
    """Round using a specific number precision using the Decimal package"""
    import warnings

    warnings.filterwarnings(action="error", category=np.ComplexWarning)
    xout = float(Decimal(float(x)).quantize(Decimal("." + precision * "0" + "1"), rounding=ROUND_HALF_EVEN))
    return xout if abs(xout) != 0.0 else 0.0


def in_ipynb():
    try:
        from IPython import get_ipython

        get_ipython()
        return True
    except NameError:
        return False


def tuple_minus(t):
    return tuple([-roundoff(x) if x != 0.0 else 0.0 for x in t])


def get_current_user():
    """Return the username of currently logged in user"""
    import getpass

    return getpass.getuser()


def thread_this(list_in, function, cpus=4):
    """
    Make a function (which only takes in a list) to run on multiple processors

    :param list_in:
    :param function:
    :param cpus:
    :return:
    """
    import multiprocessing
    from functools import partial

    var = int(len(list_in) / cpus)
    blocks = [list_in[:var]]
    for i in range(1, cpus - 1):
        blocks.append(list_in[var * i : (i + 1) * var])

    blocks.append(list_in[(cpus - 1) * var :])
    pool = multiprocessing.Pool()
    func = partial(function)
    res = pool.map(func, blocks)
    pool.close()
    pool.join()
    # Join results from the various processes
    out_res = []
    for r in res:
        out_res += r
        print(r)
    return out_res


def download_to(destination, url, file_override_ok=False):
    """

    :param destination: Destination file path
    :param url: Url of file subject for download
    :param file_override_ok: Download and write over existing file
    """
    import urllib.request

    destination = pathlib.Path(destination)
    os.makedirs(destination.parent, exist_ok=True)

    if destination.exists() and file_override_ok is False:
        print("The destination file already exists. Will skip download again")
        return

    if destination.exists() is False:
        with urllib.request.urlopen(url) as response, open(destination, "wb") as out_file:
            shutil.copyfileobj(response, out_file)


def bool2text(in_str):
    return "YES" if in_str is True else "NO"


def traverse_hdf_datasets(hdf_file):
    """Traverse all datasets across all groups in HDF5 file."""

    import h5py

    def h5py_dataset_iterator(g, prefix=""):
        for key in g.keys():
            item = g[key]
            path = "{}/{}".format(prefix, key)
            if isinstance(item, h5py.Dataset):  # test for dataset
                yield (path, item)
            elif isinstance(item, h5py.Group):  # test for group (go down)
                yield from h5py_dataset_iterator(item, path)

    with h5py.File(hdf_file, "r") as f:
        for path, dset in h5py_dataset_iterator(f):
            print(path, dset)

    return None


def zip_it(filepath):
    import pathlib
    import zipfile

    fp = pathlib.Path(filepath)
    with zipfile.ZipFile(fp.with_suffix(".zip"), "w") as zip_archive:
        zip_archive.write(fp, arcname=fp.name, compress_type=zipfile.ZIP_DEFLATED)


def zip_dir(directory, zip_path, incl_only=None):
    """

    :param directory: Directory path subject for zipping
    :param zip_path: Destination path of zip file
    :param incl_only: (optional) List of suffixes that all files should have in order to be included in the zip file
    :return:
    """

    directory = pathlib.Path(directory)

    with zipfile.ZipFile(zip_path, "w", compression=zipfile.ZIP_DEFLATED) as zip_archive:
        for root, dirs, files in os.walk(directory):
            for file in files:
                if incl_only is not None:
                    keep = False
                    suffix = file.split(".")[-1]
                    if "." + suffix in incl_only:
                        keep = True
                    if keep is False:
                        continue
                zip_archive.write(
                    os.path.join(root, file),
                    os.path.relpath(os.path.join(root, file), os.path.join(directory, "..")),
                    # compress_type=zipfile.ZIP_DEFLATED
                )


def is_within_directory(directory, target):
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return prefix == abs_directory


def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")

    tar.extractall(path, members, numeric_owner)


def unzip_it(zip_path, extract_path=None):
    fp = pathlib.Path(zip_path)
    if extract_path is None:
        extract_path = fp.parents[0]
    if fp.suffix == ".gz":
        import tarfile

        with tarfile.open(fp) as tar:
            safe_extract(tar, extract_path)
    else:
        with zipfile.ZipFile(fp, "r") as zip_archive:
            zip_archive.extractall(extract_path)


def make_name_fem_ready(value, no_dot=False):
    """
    Based on typically allowed names in FEM, this function will try to rename objects to comply without significant
    changes to the original name

    :param value:
    :param no_dot:
    :return: Fixed name
    """
    logger.debug("Converting bad name")

    if value[0] == "/":
        value = value[1:]

    value = value.replace("/", "_").replace("=", "")
    if str.isnumeric(value[0]):
        value = "_" + value

    if "/" in value:
        logger.error(f'Character "/" found in {value}')

    # if "-" in value:
    #     value = value.replace("-", "_")

    if no_dot:
        value = value.replace(".", "_")
    final_name = value.strip()
    if len(final_name) > 25:
        logger.info(f'Note FEM name "{final_name}" is >25 characters. This might cause issues in some FEM software')
    return final_name


def get_version():
    from importlib.metadata import version

    return version("ada-py")


def closest_val_in_dict(val: Union[int, float], dct: Dict[Union[int, float], Any]) -> Any:
    """
    When mapping using a dictionary and value do not match with the keys in the dictionary.
    :param val: Value a number, usually float
    :param dct: Dictionary with number keys (int o float)
    :return: Dictionary-value corresponding to the keys nearest the input value
    """
    table_looksups = np.array(list(dct))
    dct_index = table_looksups[np.abs(table_looksups - val).argmin()]
    return dct[dct_index]


def flatten(t):
    return [item for sublist in t for item in sublist]


def replace_node(old_node: Node, new_node: Node) -> None:
    from ada.fem import FemSet

    for elem in old_node.refs.copy():
        if isinstance(elem, FemSet):
            logger.warning("replace_node does not support updating FemSet")
            continue

        node_index = elem.nodes.index(old_node)

        elem.nodes.pop(node_index)
        elem.nodes.insert(node_index, new_node)
        elem.update()
        # new_node.refs.extend(old_node.refs)
        old_node.refs.pop(old_node.refs.index(elem))
        new_node.refs.append(elem)
        logger.debug(f"{old_node} exchanged with {new_node} --> {elem}")


def replace_nodes_by_tol(nodes, decimals=0, tol=Settings.point_tol):
    """

    :param nodes:
    :param decimals:
    :param tol:
    :type nodes: ada.core.containers.Nodes
    """

    def rounding(vec, decimals_):
        return np.around(vec, decimals=decimals_)

    def n_is_most_precise(n, nearby_nodes_, decimals_=0):
        most_precise = [np.array_equal(n.p, rounding(n.p, decimals_)) for n in [node] + nearby_nodes_]

        if most_precise[0] and not np.all(most_precise[1:]):
            return True
        elif not most_precise[0] and np.any(most_precise[1:]):
            return False
        elif decimals_ == 10:
            logger.error(f"Recursion started at 0 decimals, but are now at {decimals_} decimals. Will proceed with n.")
            return True
        else:
            return n_is_most_precise(n, nearby_nodes_, decimals_ + 1)

    for node in nodes:
        nearby_nodes = list(filter(lambda x: x != node, nodes.get_by_volume(node.p, tol=tol)))
        if nearby_nodes and n_is_most_precise(node, nearby_nodes, decimals):
            for nearby_node in nearby_nodes:
                replace_node(nearby_node, node)


class UnitTypes:
    LENGTH = "length"



from __future__ import annotations

import re
from dataclasses import dataclass
from itertools import chain
from typing import TYPE_CHECKING

import numpy as np

from ada.concepts.points import Node
from ada.config import get_logger
from ada.core.utils import Counter
from ada.fem import Connector, Elem
from ada.fem.containers import FemElements
from ada.fem.formats.abaqus.elem_shapes import abaqus_el_type_to_ada
from ada.fem.formats.utils import str_to_int
from ada.fem.shapes.definitions import ShapeResolver, SolidShapes

from . import cards

if TYPE_CHECKING:
    from ada.fem import FEM

_re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL
logger = get_logger()


def get_elem_from_bulk_str(bulk_str, fem: "FEM") -> FemElements:
    """Read and import all *Element flags"""
    elements = FemElements(
        chain.from_iterable(
            filter(lambda x: x is not None, (grab_elements(m, fem) for m in cards.re_el.finditer(bulk_str)))
        ),
        fem_obj=fem,
    )

    return elements


def grab_elements(match, fem: "FEM"):
    d = match.groupdict()
    eltype = d["eltype"]

    if eltype in ("CONN3D2",):
        logger.info(f'Importing Connector type "{eltype}"')

    if eltype in ("MASS", "ROTARYI"):
        logger.info(f'Importing Mass type "{eltype}"')

    ada_el_type = abaqus_el_type_to_ada(eltype)
    elset = d["elset"]
    el_type_members_str = d["members"]
    res = re.search("[a-zA-Z]", el_type_members_str)
    is_cubic = ada_el_type in [SolidShapes.HEX20, SolidShapes.HEX27]
    if is_cubic or res is None:
        if is_cubic is True:
            elem_nodes_str = el_type_members_str.splitlines()
            ntext = "".join(
                [l1.strip() + "    " + l2.strip() + "\n" for l1, l2 in zip(elem_nodes_str[:-1:2], elem_nodes_str[1::2])]
            )
        else:
            ntext = d["members"]
        res = np.fromstring(ntext.replace("\n", ","), sep=",", dtype=int)
        n = ShapeResolver.get_el_nodes_from_type(ada_el_type) + 1
        return numpy_array_to_list_of_elements(res.reshape(int(res.size / n), n), eltype, elset, ada_el_type, fem)
    else:
        elems = []
        for li in el_type_members_str.splitlines():
            elem_nodes_str = li.split(",")
            elid = str_to_int(elem_nodes_str[0])
            elem_nodes = get_elem_nodes(elem_nodes_str, fem)
            elem = Elem(elid, elem_nodes, ada_el_type, elset, el_formulation_override=eltype, parent=fem)
            elems.append(elem)
        return elems


def get_elem_nodes(elem_nodes_str, fem: "FEM"):
    elem_nodes = []
    for d in elem_nodes_str[1:]:
        temp2 = [x.strip() for x in d.split(".")]
        par_ = None
        if len(temp2) == 2:
            par, setr = temp2
            pfems = []
            parents = fem.parent.get_all_parts_in_assembly()
            for p in parents:
                pfems.append(p.fem.name)
                if p.fem.name == par:
                    par_ = p
                    break
            if par_ is None:
                raise ValueError(f'Unable to find parent for "{par}"')
            r = par_.fem.nodes.from_id(str_to_int(setr))
            if type(r) != Node:
                raise ValueError("Node ID not found")
            elem_nodes.append(r)
        else:
            r = fem.nodes.from_id(str_to_int(d))
            if type(r) != Node:
                raise ValueError("Node ID not found")
            elem_nodes.append(r)
    return elem_nodes


con_names = Counter(1, "connector")


def numpy_array_to_list_of_elements(res_, eltype, elset, ada_el_type, fem: FEM) -> list[Elem]:
    if ada_el_type == Elem.EL_TYPES.CONNECTOR_SHAPES.CONNECTOR:
        connectors = []
        for e in res_:
            if len(e) != 3:
                raise ValueError()
            el_id = e[0]
            n1, n2 = [fem.nodes.from_id(n) for n in e[1:]]
            con = Connector(next(con_names), el_id, n1, n2, con_type=None, con_sec=None, parent=fem)
            connectors.append(con)
        return connectors
    else:
        return [
            Elem(
                e[0],
                [fem.nodes.from_id(n) for n in e[1:]],
                ada_el_type,
                elset,
                el_formulation_override=eltype,
                parent=fem,
            )
            for e in res_
        ]


@dataclass
class ConnectorSectionData:
    elset: str
    behaviour: str
    connection_type: str
    csys: str


def update_connector_data(bulk_str: str, fem: FEM):
    """Extract connector elements from bulk string"""

    nsuffix = Counter(1, "_")
    for m in cards.connector_section.regex.finditer(bulk_str):
        d = m.groupdict()
        csys_ref = d["csys"].replace('"', "")
        name = d["behavior"] + next(nsuffix)
        elset = fem.elsets[d["elset"]]
        connector: Connector = elset.members[0]
        con_sec = fem.connector_sections[d["behavior"]]
        csys_ref = csys_ref[:-1] if csys_ref[-1] == "," else csys_ref
        csys = fem.lcsys[csys_ref]
        con_type = d["contype"]
        if con_type[-1] == ",":
            con_type = con_type[:-1]

        connector.elset = elset
        connector.name = name
        connector.con_sec = con_sec
        connector.con_type = con_type
        connector.csys = csys



from __future__ import annotations

import copy
from typing import TYPE_CHECKING

from ada.base.root import Root
from ada.base.units import Units

from .metals import CarbonSteel

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore


class Material(Root):
    """The base material class. Currently only supports Metals"""

    def __init__(
        self,
        name,
        mat_model=CarbonSteel("S355"),
        mat_id=None,
        parent=None,
        metadata=None,
        units=Units.M,
        guid=None,
        ifc_store: IfcStore = None,
    ):
        super(Material, self).__init__(name, guid, metadata, units, ifc_store=ifc_store)
        self._mat_model = mat_model
        mat_model.parent = self
        self._mat_id = mat_id
        self._ifc_mat = None
        self._parent = parent
        self._refs = []

    def __eq__(self, other: Material):
        """
        Assuming uniqueness of Material Name and parent

        TODO: Make this check for same Material Model parameters

        :param other:
        :return:
        """
        # other_parent = other.__dict__['_parent']
        # other_name = other.__dict__['_name']
        # if self.name == other_name and other_parent == self.parent:
        #     return True
        # else:
        #     return False

        for key, val in self.__dict__.items():
            if "parent" in key or key == "_mat_id":
                continue
            if other.__dict__[key] != val:
                return False

        return True

    def __hash__(self):
        return hash(self.guid)

    def copy_to(self, new_name: str) -> Material:
        og_parent = self.model.parent
        self.model.parent = None
        new_mat = Material(new_name, mat_model=copy.deepcopy(self.model))
        self.model.parent = og_parent
        return new_mat

    @property
    def id(self):
        return self._mat_id

    @id.setter
    def id(self, value):
        self._mat_id = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        # if value is None or any(x in value for x in [",", ".", "="]):
        #     raise ValueError(f"Material name {value} cannot be None or contain special characters")

        self._name = value.strip()

    @property
    def model(self) -> CarbonSteel:
        return self._mat_model

    @model.setter
    def model(self, value):
        value.parent = self
        self._mat_model = value

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        self.model.units = value

    @property
    def refs(self):
        return self._refs

    def __repr__(self):
        return f'Material(Name: "{self.name}" Material Model: "{self.model}'



from itertools import chain, count
from typing import Union

import numpy as np

from ada import FEM, Section
from ada.concepts.containers import Sections
from ada.config import get_logger
from ada.core.utils import roundoff
from ada.core.vector_utils import unit_vector, vector_length
from ada.fem import Csys, Elem, FemSection, FemSet
from ada.fem.containers import FemSections
from ada.fem.formats.utils import str_to_int
from ada.fem.shapes import ElemType
from ada.fem.shapes import definitions as shape_def
from ada.materials import Material
from ada.sections import GeneralProperties

from . import cards

logger = get_logger()


def get_sections(bulk_str, fem: FEM, mass_elem, spring_elem) -> FemSections:
    # Section Names
    sect_names = {sec_id: name for sec_id, name in map(get_section_names, cards.re_sectnames.finditer(bulk_str))}
    # Local Coordinate Systems
    lcsysd = {transno: vec for transno, vec in map(get_lcsys, cards.GUNIVEC.to_ff_re().finditer(bulk_str))}
    # Hinges
    hinges = {fixno: values for fixno, values in map(get_hinges, cards.re_belfix.finditer(bulk_str))}
    # Thickness'
    thick = {geono: t for geono, t in map(get_thicknesses, cards.GELTH.to_ff_re().finditer(bulk_str))}
    # Eccentricities
    ecc = {eccno: values for eccno, values in map(get_eccentricities, cards.re_geccen.finditer(bulk_str))}

    list_of_sections = chain(
        (get_isection(m, sect_names, fem) for m in cards.GIORH.to_ff_re().finditer(bulk_str)),
        (get_box_section(m, sect_names, fem) for m in cards.GBOX.to_ff_re().finditer(bulk_str)),
        (get_tubular_section(m, sect_names, fem) for m in cards.re_gpipe.finditer(bulk_str)),
        (get_flatbar(m, sect_names, fem) for m in cards.re_gbarm.finditer(bulk_str)),
    )

    fem.parent._sections = Sections(list_of_sections, parent=fem.parent)
    [add_general_sections(m, fem) for m in cards.re_gbeamg.finditer(bulk_str)]

    geom = count(1)
    total_geo = count(1)
    res = (
        get_femsecs(m, total_geo, geom, lcsysd, hinges, ecc, thick, fem, mass_elem, spring_elem)
        for m in cards.GELREF1.to_ff_re().finditer(bulk_str)
    )
    sections = filter(lambda x: type(x) is FemSection, res)

    fem_sections = FemSections(sections, fem_obj=fem)
    logger.info(f"Successfully imported {next(geom) - 1} FEM sections out of {next(total_geo) - 1}")
    return fem_sections


def get_isection(match, sect_names, fem) -> Section:
    d = match.groupdict()
    sec_id = str_to_int(d["geono"])
    name = sect_names[sec_id]
    return Section(
        name=name,
        sec_id=sec_id,
        sec_type=Section.TYPES.IPROFILE,
        h=roundoff(d["hz"]),
        t_w=roundoff(d["ty"]),
        w_top=roundoff(d["bt"]),
        w_btn=roundoff(d["bb"]),
        t_ftop=roundoff(d["tt"]),
        t_fbtn=roundoff(d["tb"]),
        parent=fem.parent,
    )


def get_box_section(match, sect_names, fem) -> Section:
    d = match.groupdict()
    sec_id = str_to_int(d["geono"])
    return Section(
        name=sect_names[sec_id],
        sec_id=sec_id,
        sec_type=Section.TYPES.BOX,
        h=roundoff(d["hz"]),
        w_top=roundoff(d["by"]),
        w_btn=roundoff(d["by"]),
        t_w=roundoff(d["ty"]),
        t_ftop=roundoff(d["tt"]),
        t_fbtn=roundoff(d["tb"]),
        parent=fem.parent,
    )


def get_flatbar(match, sect_names, fem) -> Section:
    d = match.groupdict()
    sec_id = str_to_int(d["geono"])
    return Section(
        name=sect_names[sec_id],
        sec_id=sec_id,
        sec_type=Section.TYPES.FLATBAR,
        h=roundoff(d["hz"]),
        w_top=roundoff(d["bt"]),
        w_btn=roundoff(d["bb"]),
        parent=fem.parent,
    )


def add_general_sections(match, fem) -> None:
    d = match.groupdict()
    sec_id = str_to_int(d["geono"])
    gen_props = GeneralProperties(
        Ax=roundoff(d["area"], 10),
        Ix=roundoff(d["ix"], 10),
        Iy=roundoff(d["iy"], 10),
        Iz=roundoff(d["iz"], 10),
        Iyz=roundoff(d["iyz"], 10),
        Wxmin=roundoff(d["wxmin"]),
        Wymin=roundoff(d["wymin"]),
        Wzmin=roundoff(d["wzmin"]),
        Shary=roundoff(d["shary"]),
        Sharz=roundoff(d["sharz"]),
        Shceny=roundoff(d["shceny"]),
        Shcenz=roundoff(d["shcenz"]),
        Sy=float(d["sy"]),
        Sz=float(d["sz"]),
    )

    if sec_id in fem.parent.sections.id_map.keys():
        sec = fem.parent.sections.get_by_id(sec_id)
        sec._genprops = gen_props
        gen_props.parent = sec
    else:
        stype = Section.TYPES.GENERAL
        sec = Section(name=f"GB{sec_id}", sec_id=sec_id, sec_type=stype, genprops=gen_props, parent=fem.parent)
        gen_props.parent = sec
        fem.parent.sections.add(sec)


def get_tubular_section(match, sect_names, fem) -> Section:
    d = match.groupdict()
    sec_id = str_to_int(d["geono"])
    if sec_id not in sect_names:
        sec_name = f"TUB{sec_id}"
    else:
        sec_name = sect_names[sec_id]
    t = d["t"] if d["t"] is not None else (d["dy"] - d["di"]) / 2
    return Section(
        name=sec_name,
        sec_id=sec_id,
        sec_type=Section.TYPES.TUBULAR,
        r=roundoff(float(d["dy"]) / 2),
        wt=roundoff(t),
        parent=fem.parent,
    )


def read_line_section(elem: Elem, fem: FEM, mat: Material, geono, d, lcsysd, hinges_global, eccentricities):
    transno = str_to_int(d["transno"])
    sec = fem.parent.sections.get_by_id(geono)
    n1, n2 = elem.nodes
    v = n2.p - n1.p
    if vector_length(v) == 0.0:
        xvec = [1, 0, 0]
    else:
        xvec = unit_vector(v)
    zvec = lcsysd[transno]
    crossed = np.cross(zvec, xvec)
    ma = max(abs(crossed))
    yvec = tuple([roundoff(x / ma, 3) for x in crossed])

    fix_data = str_to_int(d["fixno"])
    ecc_data = str_to_int(d["eccno"])

    members = None
    if d["members"] is not None:
        members = [str_to_int(x) for x in d["members"].replace("\n", " ").split()]

    if fix_data == -1:
        add_hinge_prop_to_elem(elem, members, hinges_global, xvec, yvec)

    if ecc_data == -1:
        add_ecc_to_elem(elem, members, eccentricities, fix_data)

    fem_set = FemSet(sec.name, [elem], "elset", metadata=dict(internal=True), parent=fem)
    fem.sets.add(fem_set, append_suffix_on_exist=True)
    fem_sec = FemSection(
        sec_id=geono,
        name=sec.name,
        sec_type=ElemType.LINE,
        elset=fem_set,
        section=sec,
        local_z=zvec,
        local_y=yvec,
        material=mat,
        parent=fem,
    )
    return fem_sec


def read_shell_section(elem: Elem, fem: FEM, mat: Material, elno, thicknesses, geono):
    sec_name = f"sh{elno}"
    fem_set = FemSet(sec_name, [elem], "elset", parent=fem, metadata=dict(internal=True))
    fem.sets.add(fem_set)
    fem_sec = FemSection(
        name=sec_name,
        sec_type=ElemType.SHELL,
        thickness=roundoff(thicknesses[geono]),
        elset=fem_set,
        material=mat,
        parent=fem,
    )
    return fem_sec


def get_femsecs(match, total_geo, curr_geom_num, lcsysd, hinges_global, ecc, thicknesses, fem, mass_elem, spring_elem):
    next(total_geo)
    d = match.groupdict()
    geono = str_to_int(d["geono"])
    elno = str_to_int(d["elno"])
    matno = str_to_int(d["matno"])

    # Go no further if element has no fem section
    if elno in spring_elem.keys():
        next(curr_geom_num)
        spring_elem[elno]["section_data"] = d
        return None

    if elno in mass_elem.keys():
        next(curr_geom_num)
        mass_elem[elno]["section_data"] = d
        return None

    elem = fem.elements.from_id(elno)
    mat = fem.parent.materials.get_by_id(matno)
    if isinstance(elem.type, shape_def.LineShapes):
        next(curr_geom_num)
        return read_line_section(elem, fem, mat, geono, d, lcsysd, hinges_global, ecc)
    elif isinstance(elem.type, shape_def.ShellShapes):
        next(curr_geom_num)
        return read_shell_section(elem, fem, mat, elno, thicknesses, geono)
    else:
        raise ValueError("Section not added to conversion")


def get_thicknesses(match):
    d = match.groupdict()
    sec_id = str_to_int(d["geono"])
    t = d["th"]
    return sec_id, t


def get_hinges(match):
    d = match.groupdict()
    fixno = str_to_int(d["fixno"])
    opt = str_to_int(d["opt"])
    trano = str_to_int(d["trano"])
    a1 = str_to_int(d["a1"])
    a2 = str_to_int(d["a2"])
    a3 = str_to_int(d["a3"])
    a4 = str_to_int(d["a4"])
    a5 = str_to_int(d["a5"])
    try:
        a6 = str_to_int(d["a6"])
    except BaseException as e:
        logger.debug(e)
        a6 = 0
        pass
    return fixno, (opt, trano, a1, a2, a3, a4, a5, a6)


def get_eccentricities(match):
    d = match.groupdict()
    eccno = str_to_int(d["eccno"])
    ex = float(d["ex"])
    ey = float(d["ey"])
    ez = float(d["ez"])
    return eccno, (ex, ey, ez)


def get_section_names(m):
    d = m.groupdict()
    return str_to_int(d["geono"]), d["set_name"].strip()


def add_hinge_prop_to_elem(elem: Elem, members, hinges_global, xvec, yvec) -> None:
    """Add hinge property to element from sesam FEM file"""
    from ada.fem.elements import Hinge, HingeProp

    if len(elem.nodes) > 2:
        raise ValueError("This algorithm was not designed for more than 2 noded elements")

    for i, x in enumerate(members):
        if i >= len(elem.nodes):
            break
        if x == 0:
            continue
        if x not in hinges_global.keys():
            raise ValueError("fixno not found!")
        opt, trano, a1, a2, a3, a4, a5, a6 = hinges_global[x]
        n = elem.nodes[i]
        if trano > 0:
            csys = None
        else:
            csys = Csys(
                f"el{elem.id}_hinge{i + 1}_csys",
                coords=([unit_vector(xvec) + n.p, unit_vector(yvec) + n.p, n.p]),
                parent=elem.parent,
            )
        dofs_origin = [1, 2, 3, 4, 5, 6]
        dofs = [int(x) for x, i in zip(dofs_origin, (a1, a2, a3, a4, a5, a6)) if int(i) != 0]
        end = Hinge(retained_dofs=dofs, csys=csys, fem_node=n)
        if i == 0:
            elem.hinge_prop = HingeProp(end1=end)
        else:
            elem.hinge_prop = HingeProp(end2=end)


def add_ecc_to_elem(elem: Elem, members, eccentricities, fix_data) -> None:
    """Adds eccentricity to element from sesam FEM file"""
    from ada.fem.elements import Eccentricity, EccPoint

    # To the interpretation here
    start = 0 if fix_data != -1 else len(elem.nodes)
    end = len(elem.nodes) - 1 if fix_data != -1 else 2 * len(elem.nodes)
    end1: Union[None, EccPoint] = None
    end2: Union[None, EccPoint] = None
    for i, x in enumerate(members[start:]):
        if i > end:
            break
        if x == 0:
            continue
        n_offset = elem.nodes[i]
        ecc = eccentricities[x]
        if i == 0:
            end1 = EccPoint(n_offset, ecc)
        if i == end:
            end2 = EccPoint(n_offset, ecc)

    elem.eccentricity = Eccentricity(end1, end2)


def get_lcsys(m):
    d = m.groupdict()
    return str_to_int(d["transno"]), (
        roundoff(d["unix"]),
        roundoff(d["uniy"]),
        roundoff(d["uniz"]),
    )



import os
from dataclasses import dataclass


@dataclass
class User:
    user_id: str = os.environ.get("ADAUSER", "AdaUser")
    given_name: str = None
    family_name: str = None
    middle_names: str = None
    prefix_titles: str = None
    suffix_titles: str = None
    org_id: str = "ADA"
    org_name: str = "Assembly For Design and Analysis"
    org_description: str = None
    role: str = "Engineer"
    parent = None



from __future__ import annotations

from typing import TYPE_CHECKING

from ada.fem.exceptions.element_support import IncompatibleElements
from ada.fem.formats.utils import get_fem_model_from_assembly
from ada.fem.shapes.definitions import ElemType

if TYPE_CHECKING:
    from ada import Assembly


def check_compatibility(assembly: Assembly):
    p = get_fem_model_from_assembly(assembly)
    step = assembly.fem.steps[0] if len(assembly.fem.steps) > 0 else None

    if step is not None:
        has_nonlin = True if True in [x.nl_geom for x in assembly.fem.steps] else False
        for el in p.fem.elements.lines:
            if has_nonlin:
                raise IncompatibleElements(
                    "The standard Euler/Timoshenko beams in Code Aster do not support nonlinear"
                    " material assignment. todo: add option to auto-assign beams linear"
                    "material, and/or add support for fiber beams"
                )
            if el.type == ElemType.LINE_SHAPES.LINE3:
                raise IncompatibleElements("2nd order beam elements are currently not supported in Code Aster")

        for el in p.fem.elements.shell:
            if el.type in (ElemType.SHELL_SHAPES.QUAD8, ElemType.SHELL_SHAPES.TRI6) and has_nonlin:
                raise IncompatibleElements(
                    "The default 2nd order shell elements are currently not supported in "
                    "Code Aster when running the analysis using nonlinear materials"
                )



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING

from .helper_utils import get_instance_name

if TYPE_CHECKING:
    from ada import FEM
    from ada.fem import Csys


def orientations_str(fem: FEM, written_on_assembly_level: bool) -> str:
    """Add orientations associated with loads"""
    cstr = "** Orientations associated with Loads"
    for step in fem.steps:
        for load in step.loads:
            if load.csys is None:
                continue
            cstr += "\n"
            coord_str = ", ".join([str(x) for x in chain.from_iterable(load.csys.coords)])[:-1]
            name = load.fem_set.name.upper()
            inst_name = get_instance_name(load.fem_set, written_on_assembly_level)
            cstr += f"*Nset, nset=_T-{name}, internal\n{inst_name},\n"
            cstr += f"*Transform, nset=_T-{name}\n{coord_str}\n"
            cstr += csys_str(load.csys, written_on_assembly_level)

    return cstr.strip()


def csys_str(csys: Csys, written_on_assembly_level: bool):
    """"""
    name = csys.name

    def f(num: float) -> str:
        return f"{num:.3f}"

    ori_str = f'*Orientation, name="{name}"'
    if csys.nodes is None and csys.coords is None:
        ori_str += "\n 1.,           0.,           0.,           0.,           1.,           0.\n 1, 0."
    elif csys.nodes is not None:
        if len(csys.nodes) != 3:
            raise ValueError("CSYS number of nodes must be 3")
        ori_str += ", DEFINITION=NODES, SYSTEM=RECTANGULAR\n {},{},{}".format(
            *[get_instance_name(no, written_on_assembly_level) for no in csys.nodes]
        )
    else:
        ax, ay, az = csys.coords[0]
        ori_str += f" \n {f(ax)}, {f(ay)}, {f(az)}"
        bx, by, bz = csys.coords[1]
        ori_str += f", {f(bx)}, {f(by)}, {f(bz)}"
        if len(csys.coords) == 3:
            cx, cy, cz = csys.coords[2]
            ori_str += f", {f(cx)}, {f(cy)}, {f(cz)}"
        ori_str += "\n 1, 0."
    return ori_str



from typing import TYPE_CHECKING

import numpy as np

from ada import CurvePoly

if TYPE_CHECKING:
    from ada import Beam

    from ..concepts import GmshData, GmshSession


def ibeam(model: "GmshData", gmsh_session: "GmshSession"):
    # pass
    bm_obj: "Beam" = model.obj
    for cut in make_ig_cutplanes(bm_obj):
        gmsh_session.add_cutting_plane(cut, [model])

    gmsh_session.make_cuts()

    for dim, tag in gmsh_session.model.get_entities():
        if dim == 2:
            gmsh_session.model.mesh.set_transfinite_surface(tag)
            gmsh_session.model.mesh.setRecombine(dim, tag)

    for dim, tag in gmsh_session.model.get_entities():
        if dim == 3:
            gmsh_session.model.mesh.set_transfinite_volume(tag)
            gmsh_session.model.mesh.setRecombine(dim, tag)

    # gmsh_session.open_gui()

    # raise NotImplementedError()


def ar(*iterable):
    """Short form for creating numpy array"""
    r = np.array(iterable[0])
    return r


def make_ig_cutplanes(bm: "Beam"):
    from ..concepts import CutPlane

    bm1_sec_curve = get_bm_section_curve(bm)
    minz = min([x[2] for x in bm1_sec_curve.points3d])
    maxz = max([x[2] for x in bm1_sec_curve.points3d])
    pmin, pmax = bm.bbox().p1, bm.bbox().p2
    dx, dy, dz = (ar(pmax) - ar(pmin)) * 1.0
    x, y, _ = pmin

    sec = bm.section

    cut1 = CutPlane((x, y, minz + sec.t_fbtn), dx=dx, dy=dy)
    cut2 = CutPlane((x, y, maxz - sec.t_fbtn), dx=dx, dy=dy)

    web_left = bm.n1.p - (sec.t_w / 2) * bm.yvec - (sec.h / 2) * bm.up
    web_right = bm.n1.p + (sec.t_w / 2) * bm.yvec - (sec.h / 2) * bm.up
    dy = sec.h
    cut3 = CutPlane(web_left, dx=dx, dy=dy, plane="XZ")
    cut4 = CutPlane(web_right, dx=dx, dy=dy, plane="XZ")

    return [cut1, cut2, cut3, cut4]


def get_bm_section_curve(bm: "Beam", origin=None) -> CurvePoly:
    origin = origin if origin is not None else bm.n1.p
    section_profile = bm.section.get_section_profile(True)
    points2d = section_profile.outer_curve.points2d
    return CurvePoly(points2d=points2d, origin=origin, xdir=bm.yvec, normal=bm.xvec, parent=bm.parent)



import pathlib
import re

from ada.config import get_logger
from ada.fem.formats.sesam.results.sin2sif import convert_sin_to_sif
from ada.fem.formats.utils import LocalExecute

logger = get_logger()


def run_sesam(
    inp_path: pathlib.Path,
    cpus=2,
    gpus=None,
    run_ext=False,
    metadata=None,
    execute=True,
    exit_on_complete=True,
    run_in_shell=False,
):
    logger.info("sestra runs on single core only. changing cpus=1")
    cpus = 1
    ses_exe = SesamExecute(
        inp_path, cpus=cpus, run_ext=run_ext, metadata=metadata, auto_execute=execute, run_in_shell=run_in_shell
    )
    out = ses_exe.run(exit_on_complete)
    reg_res = re.search(r"T([0-9]{0,4})\.FEM", inp_path.name)
    tbr = reg_res.group()
    num = reg_res.group(1)

    sin_file = inp_path.parent / inp_path.name.replace(tbr, f"R{num}.SIN")
    convert_sin_to_sif(sin_file)

    return out


class SesamExecute(LocalExecute):
    def run(self, exit_on_complete=True, run_cmd=None, bat_start_str=None):
        from ada.fem.formats.general import FEATypes

        exe_path = self.get_exe(FEATypes.SESAM)
        if run_cmd is None:
            run_cmd = f"{exe_path} /dsf {self.analysis_name}T100"
        stop_cmd = None
        out = self._run_local(run_cmd, stop_cmd, exit_on_complete, bat_start_str)
        return out



import numpy as np

from ada.config import get_logger

logger = get_logger()


def bernstein(n, k):
    """Bernstein polynomial."""
    from scipy.special._ufuncs import binom

    coeff = binom(n, k)

    def _bpoly(x):
        return coeff * x**k * (1 - x) ** (n - k)

    return _bpoly


def bezier(points, num=200):
    """Build Bezier curve from points."""
    N = len(points)
    t = np.linspace(0, 1, num=num)
    curve = np.zeros((num, 2))
    for ii in range(N):
        curve += np.outer(bernstein(N - 1, ii)(t), points[ii])
    return curve


def curve_fitting(in_data):
    from scipy.optimize import curve_fit

    xData = np.array(in_data[0])
    yData = np.array(in_data[1])

    # generate initial parameter values
    geneticParameters = generate_initial_parameters(xData, yData)

    # curve fit the test data
    fittedParameters, pcov = curve_fit(curve_f1, xData, yData, geneticParameters)

    logger.debug("Parameters", fittedParameters)

    modelPredictions = curve_f1(xData, *fittedParameters)

    absError = modelPredictions - yData

    SE = np.square(absError)  # squared errors
    MSE = np.mean(SE)  # mean squared errors
    RMSE = np.sqrt(MSE)  # Root Mean Squared Error, RMSE
    Rsquared = 1.0 - (np.var(absError) / np.var(yData))
    print("RMSE:", RMSE)
    print("R-squared:", Rsquared)
    print()
    return fittedParameters


def curve_f1(x, a, b, Offset):
    """A base function for use in curve fitting"""
    return 1.0 / (1.0 + np.exp(-a * (x - b))) + Offset


def curve_f2(x, a, b, c):
    """Another base function for use in curve fitting"""
    return a * np.exp(-b * x) + c


def curve_f3(x, a, b):
    """Yet another base function for use in curve fitting"""
    return a * np.exp(b * x)


def curve_f4(x, a, b, c):
    """And yet another base function for use in curve fitting"""
    return a * x**3 + b * x**2 + c * x


def sum_of_squared_error(parameter_tuple, *args):
    """function for genetic algorithm to minimize (sum of squared error)"""
    import warnings

    xData = args[0]
    yData = args[1]
    if xData is None:
        logger.error("Xdata is none. Returning None")
        return None
    warnings.filterwarnings("ignore")
    val = curve_f1(xData, *parameter_tuple)
    return np.sum((yData - val) ** 2.0)


def generate_initial_parameters(xData, yData):
    from scipy.optimize import differential_evolution

    # min and max used for bounds
    maxX = max(xData)
    minX = min(xData)
    maxY = max(yData)
    # minY = min(yData)

    parameterBounds = []
    parameterBounds.append([minX, maxX])  # seach bounds for a
    parameterBounds.append([minX, maxX])  # seach bounds for b
    parameterBounds.append([0.0, maxY])  # seach bounds for Offset

    # "seed" the numpy random number generator for repeatable results
    result = differential_evolution(sum_of_squared_error, parameterBounds, args=[xData, yData], seed=3)
    return result.x



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Iterable

import numpy as np

from ada.base.physical_objects import BackendGeom
from ada.base.types import GeomRepr
from ada.config import get_logger
from ada.core.utils import thread_this
from ada.occ.exceptions.geom_creation import (
    UnableToBuildNSidedWires,
    UnableToCreateSolidOCCGeom,
    UnableToCreateTesselationFromSolidOCCGeom,
)
from ada.visualize.concept import ObjectMesh
from ada.visualize.config import ExportConfig
from ada.visualize.renderer_occ import occ_shape_to_faces

if TYPE_CHECKING:
    from ada import Beam, PipeSegElbow, PipeSegStraight, Plate, Shape, Wall

logger = get_logger()


def filter_mesh_objects(objects: Iterable[BackendGeom], export_config: ExportConfig) -> None | list[BackendGeom]:
    from ada import Pipe

    guid_filter = export_config.data_filter.filter_elements_by_guid
    obj_list: list[BackendGeom] = []

    for obj in objects:
        if guid_filter is not None and obj.guid not in guid_filter:
            continue
        if isinstance(obj, Pipe):
            for seg in obj.segments:
                obj_list.append(seg)
        else:
            obj_list.append(obj)

    if len(obj_list) == 0:
        return None

    return obj_list


def ifc_poly_elem_to_json(
    obj: Shape, export_config: ExportConfig = ExportConfig(), opt_func: Callable = None
) -> list[ObjectMesh]:
    import ifcopenshell.geom

    from ada.ifc.utils import create_guid, get_representation_items
    from ada.visualize.utils import merge_mesh_objects, organize_by_colour

    a = obj.get_assembly()
    ifc_store = a.ifc_store
    ifc_f = ifc_store.f
    ifc_elem = ifc_f.by_guid(obj.guid)

    settings = ifcopenshell.geom.settings()
    settings.set(settings.USE_PYTHON_OPENCASCADE, False)
    settings.set(settings.SEW_SHELLS, False)
    settings.set(settings.WELD_VERTICES, True)
    settings.set(settings.INCLUDE_CURVES, False)
    settings.set(settings.USE_WORLD_COORDS, True)
    settings.set(settings.VALIDATE_QUANTITIES, False)

    geom_items = get_representation_items(ifc_f, ifc_elem)
    meshes = []
    if len(geom_items) == 0:
        raise ValueError("No IFC geometries found.")

    for i, ifc_geom in enumerate(geom_items):
        geometry = ifc_store.get_ifc_geom(ifc_geom, settings)
        guid = obj.guid if i == 0 else create_guid()
        vertices = np.array(geometry.verts, dtype="float32").reshape(int(len(geometry.verts) / 3), 3)
        faces = np.array(geometry.faces, dtype=int)
        normals = np.array(geometry.normals) if len(geometry.normals) != 0 else None

        if normals is not None and len(normals) > 0:
            normals = normals.astype(dtype="float32").reshape(int(len(normals) / 3), 3)

        if opt_func is not None:
            faces, vertices, normals = opt_func(faces.reshape(int(len(geometry.faces) / 3), 3), vertices, normals)
            vertices = vertices.astype(dtype="float32").flatten()
            faces = faces.astype(dtype="int32").flatten()
            if normals is not None:
                normals = normals.astype(dtype="float32").flatten()

        mats = geometry.materials
        if len(mats) == 0:
            colour = [1.0, 0.0, 0.0, 1.0]
        else:
            mat0 = mats[0]
            opacity = 1.0 - mat0.transparency
            colour = [*mat0.diffuse, opacity]

        obj_mesh = ObjectMesh(guid, faces, vertices, normals, colour, translation=export_config.volume_center)
        meshes.append(obj_mesh)

    if export_config.merge_subgeometries_by_colour is False or len(meshes) == 1:
        return meshes

    # merge subgeometries by colour
    colour_map = organize_by_colour(meshes)
    merged_meshes = []
    main_obj = False
    applied_guid = False
    for i, (colour, elements) in enumerate(colour_map.items()):
        obj_mesh = merge_mesh_objects(elements)

        if len(obj_mesh.faces) == 0:
            continue

        if colour[-1] == 1.0 and applied_guid is False:
            main_obj = True
            obj_mesh.guid = obj.guid
            applied_guid = True

        merged_meshes.append(obj_mesh)

    if main_obj is False:
        merged_meshes[0].guid = obj.guid

    return merged_meshes


def occ_geom_to_poly_mesh(
    obj: BackendGeom,
    export_config: ExportConfig = ExportConfig(),
    opt_func: Callable = None,
    geom_repr: GeomRepr = GeomRepr.SOLID,
) -> ObjectMesh:
    if geom_repr == GeomRepr.SOLID:
        geom = obj.solid()
    elif geom_repr == GeomRepr.SHELL:
        geom = obj.shell()
    else:
        export_config.render_edges = True
        geom = obj.line()

    position, indices, normals, _ = occ_shape_to_faces(
        geom,
        export_config.quality,
        export_config.render_edges,
        export_config.parallel,
    )

    if opt_func is not None:
        indices, position, normals = opt_func(indices, position, normals)
    else:
        opt_func_example(indices, position, normals)

    colour = [*obj.colour_norm, obj.opacity]
    return ObjectMesh(obj.guid, indices, position, normals, colour, translation=export_config.volume_center)


def obj_to_mesh(
    obj: Beam | Plate | Wall | PipeSegElbow | PipeSegStraight | Shape,
    export_config: ExportConfig = ExportConfig(),
    opt_func: Callable = None,
) -> None | list[ObjectMesh]:
    if obj.parent.get_assembly().ifc_store is not None and export_config.ifc_skip_occ is True:
        try:
            obj_meshes = ifc_poly_elem_to_json(obj, export_config, opt_func)
        except RuntimeError as e:
            logger.error(e)
            return None
    else:
        try:
            obj_meshes = occ_geom_to_poly_mesh(obj, export_config, opt_func)
        except (UnableToBuildNSidedWires, UnableToCreateTesselationFromSolidOCCGeom, UnableToCreateSolidOCCGeom) as e:
            logger.error(e)
            return None

    return obj_meshes


def id_map_using_threading(list_in, threads: int):
    # obj = list_in[0]
    # obj_str = json.dumps(obj)
    # serialize_evaluator(obj)
    res = thread_this(list_in, obj_to_mesh, threads)
    print(res)
    return res


def opt_func_example(faces, position, normals):
    """Optimize by finding removing vertices with same coordinates and normals"""
    obj_buffer_arrays = np.concatenate([position, normals], 1)
    buffer, indices = np.unique(obj_buffer_arrays, axis=0, return_index=False, return_inverse=True)
    x, y, z, nx, ny, nz = buffer.T
    position = np.array([x, y, z]).T
    normals = np.array([nx, ny, nz]).T
    return faces, position, normals



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import h5py
import numpy as np

from ada.concepts.containers import Nodes
from ada.config import get_logger
from ada.fem import Elem
from ada.fem.containers import FemElements, FemSets

from ..common import med_to_ada_type
from .read_sets import (
    _cell_tag_to_set,
    _element_set_dict_to_list_of_femset,
    _point_tags_to_sets,
    _read_families,
)

if TYPE_CHECKING:
    from ada.concepts.spatial import Assembly
    from ada.fem import FEM

logger = get_logger()


def read_fem(fem_file: os.PathLike, fem_name: str = None) -> Assembly:
    from ada import Assembly, Part

    fem = med_to_fem(fem_file, fem_name)
    if fem_name is None:
        fem_name = fem.name

    return Assembly("TempAssembly") / Part(fem_name, fem=fem)


def med_to_fem(fem_file, fem_name) -> FEM:
    from ada import FEM, Node

    with h5py.File(fem_file, "r") as f:
        # Mesh ensemble
        mesh_ensemble = f["ENS_MAA"]
        meshes = mesh_ensemble.keys()
        if len(meshes) != 1:
            raise ValueError("Must only contain exactly 1 mesh, found {}.".format(len(meshes)))

        mesh_name = list(meshes)[0]
        mesh = mesh_ensemble[mesh_name]

        dim = mesh.attrs["ESP"]

        fem_name = fem_name if fem_name is not None else mesh_name

        # Initialize FEM object
        fem = FEM(fem_name)

        # Possible time-stepping
        if "NOE" not in mesh:
            # One needs NOE (node) and MAI (French maillage, meshing) data. If they
            # are not available in the mesh, check for time-steppings.
            time_step = mesh.keys()
            if len(time_step) != 1:
                raise ValueError(f"Must only contain exactly 1 time-step, found {len(time_step)}.")
            mesh = mesh[list(time_step)[0]]

        # Points
        pts_dataset = mesh["NOE"]["COO"]
        n_points = pts_dataset.attrs["NBR"]
        points = pts_dataset[()].reshape((n_points, dim), order="F")

        if "NUM" in mesh["NOE"]:
            point_num = list(mesh["NOE"]["NUM"])
        else:
            logger.warning("No node information is found on MED file")
            point_num = np.arange(1, len(points) + 1)

        fem.nodes = Nodes([Node(p, point_num[i], parent=fem) for i, p in enumerate(points)], parent=fem)

        # Point tags
        tags = None
        if "FAM" in mesh["NOE"]:
            tags = mesh["NOE"]["FAM"][()]

        # Information for point tags
        point_tags = {}
        fas = mesh["FAS"] if "FAS" in mesh else f["FAS"][mesh_name]
        if "NOEUD" in fas:
            point_tags = _read_families(fas["NOEUD"])

        point_sets = _point_tags_to_sets(tags, point_tags, fem) if tags is not None else []

        # Information for cell tags
        cell_tags = {}
        if "ELEME" in fas:
            cell_tags = _read_families(fas["ELEME"])

        # CellBlock
        cell_types = []
        med_cells = mesh["MAI"]

        elements = []
        element_sets = dict()
        for med_cell_type, med_cell_type_group in med_cells.items():
            if med_cell_type == "PO1":
                logger.warning("Point elements are still not supported")
                continue

            cell_type = med_to_ada_type(med_cell_type)
            cell_types.append(cell_type)

            nod = med_cell_type_group["NOD"]
            n_cells = nod.attrs["NBR"]
            nodes_in = nod[()].reshape(n_cells, -1, order="F")

            if "NUM" in med_cell_type_group.keys():
                num = list(med_cell_type_group["NUM"])
            else:
                num = np.arange(0, len(nodes_in))

            element_block = [
                Elem(num[i], [fem.nodes.from_id(e) for e in c], cell_type, parent=fem) for i, c in enumerate(nodes_in)
            ]
            elements += element_block
            # Cell tags
            if "FAM" in med_cell_type_group:
                cell_data = med_cell_type_group["FAM"][()]
                cell_type_sets = _cell_tag_to_set(cell_data, cell_tags)
                for key, val in cell_type_sets.items():
                    if key not in element_sets.keys():
                        element_sets[key] = []
                    # TODO: set(val) returns unordered list
                    element_sets[key] += [element_block[i] for i in set(val)]

    fem.elements = FemElements(elements, fem_obj=fem)

    elsets = _element_set_dict_to_list_of_femset(element_sets, fem)

    fem.sets = FemSets(elsets + point_sets, parent=fem)
    return fem


def _read_nodal_data(med_data, profiles):
    profile = med_data["NOE"].attrs["PFL"]
    data_profile = med_data["NOE"][profile]
    n_points = data_profile.attrs["NBR"]
    if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
        values = data_profile["CO"][()].reshape(n_points, -1, order="F")
    else:
        n_data = profiles[profile].attrs["NBR"]
        index_profile = profiles[profile]["PFL"][()] - 1
        values_profile = data_profile["CO"][()].reshape(n_data, -1, order="F")
        values = np.full((n_points, values_profile.shape[1]), np.nan)
        values[index_profile] = values_profile
    if values.shape[-1] == 1:  # cut off for scalars
        values = values[:, 0]
    return values


def _read_cell_data(med_data, profiles):
    profile = med_data.attrs["PFL"]
    data_profile = med_data[profile]
    n_cells = data_profile.attrs["NBR"]
    n_gauss_points = data_profile.attrs["NGA"]
    if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
        values = data_profile["CO"][()].reshape(n_cells, n_gauss_points, -1, order="F")
    else:
        n_data = profiles[profile].attrs["NBR"]
        index_profile = profiles[profile]["PFL"][()] - 1
        values_profile = data_profile["CO"][()].reshape(n_data, n_gauss_points, -1, order="F")
        values = np.full((n_cells, values_profile.shape[1], values_profile.shape[2]), np.nan)
        values[index_profile] = values_profile

    # Only 1 data point per cell, shape -> (n_cells, n_components)
    if n_gauss_points == 1:
        values = values[:, 0, :]
        if values.shape[-1] == 1:  # cut off for scalars
            values = values[:, 0]
    return values



from __future__ import annotations

import os
import pathlib

import ifcopenshell
import ifcopenshell.draw as draw


def create_svg_from_drawings(files: list[str], svg_output: str | pathlib.Path):
    settings = draw.draw_settings()
    if isinstance(svg_output, str):
        svg_output = pathlib.Path(svg_output)

    result = draw.main(settings, [ifcopenshell.open(f) for f in files], merge_projection=True)

    os.makedirs(svg_output.parent, exist_ok=True)
    with open(svg_output, "wb") as f:
        f.write(result)



from __future__ import annotations

import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

from ada.core.constants import X, Y, Z

from .write_utils import add_local_system

if TYPE_CHECKING:
    from ada import Part


def add_boundary_conditions(root: ET.Element, part: Part):
    dof_map = {y: x for x, y in dict(dx=1, dy=2, dz=3, rx=4, ry=5, rz=6).items()}

    all_bc = list(part.fem.get_all_bcs())
    for bc in all_bc:
        if len(bc.fem_set.members) != 1:
            raise NotImplementedError()

        n = bc.fem_set.members[0]

        bc_stru = ET.SubElement(root, "structure")
        sup_point = ET.SubElement(bc_stru, "support_point", {"name": bc.name})
        sup_point.append(add_local_system(X, Y, Z))
        geom = ET.SubElement(sup_point, "geometry")
        ET.SubElement(geom, "position", {"x": str(n.x), "y": str(n.y), "z": str(n.z)})
        bc_con = ET.SubElement(sup_point, "boundary_conditions")
        for dof in range(1, 7):
            ftyp = "fixed" if dof in bc.dofs else "free"
            ET.SubElement(bc_con, "boundary_condition", dict(constraint=ftyp, dof=dof_map.get(dof)))



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import ifcopenshell

from ada import Material
from ada.config import get_logger

if TYPE_CHECKING:
    from ada.ifc.store import IfcStore

logger = get_logger()


def read_material(ifc_mat: ifcopenshell.entity_instance, ifc_store: IfcStore) -> Material:
    mi = MaterialImporter(ifc_store)
    return mi.read_material(ifc_mat)


@dataclass
class MaterialImporter:
    ifc_store: IfcStore

    def load_ifc_materials(self):
        for ifc_mat in self.ifc_store.f.by_type("IfcMaterial"):
            mat = self.ifc_store.assembly.add_material(self.read_material(ifc_mat))
            logger.info(f'Importing material "{mat}"')

    def read_material(self, ifc_mat: ifcopenshell.entity_instance) -> Material:
        from ada.materials.metals import CarbonSteel, Metal

        mat_psets = ifc_mat.HasProperties if hasattr(ifc_mat, "HasProperties") else None

        if mat_psets is None or len(mat_psets) == 0:
            logger.info(f'No material properties found for "{ifc_mat}"')
            return Material(ifc_mat.Name)

        props = {}
        for entity in mat_psets[0].Properties:
            if entity.is_a("IfcPropertySingleValue"):
                props[entity.Name] = entity.NominalValue[0]

        mat_props = dict(
            E=props.get("YoungModulus", 210000e6),
            sig_y=props.get("YieldStress", 355e6),
            rho=props.get("MassDensity", 7850),
            v=props.get("PoissonRatio", 0.3),
            alpha=props.get("ThermalExpansionCoefficient", 1.2e-5),
            zeta=props.get("SpecificHeatCapacity", 1.15),
            units=self.ifc_store.assembly.units,
        )

        if "StrengthGrade" in props:
            mat_model = CarbonSteel(grade=props["StrengthGrade"], **mat_props)
        else:
            mat_model = Metal(sig_u=None, **mat_props)

        guid = None
        if len(ifc_mat.AssociatedTo) == 1:
            guid = ifc_mat.AssociatedTo[0].GlobalId

        return Material(
            name=ifc_mat.Name,
            mat_model=mat_model,
            ifc_store=self.ifc_store,
            units=self.ifc_store.assembly.units,
            guid=guid,
        )



from itertools import chain

import numpy as np

from ada.config import get_logger
from ada.core.utils import roundoff
from ada.fem import FEM, Elem, FemSet, Mass, Spring
from ada.fem.containers import FemElements
from ada.fem.formats.sesam.common import sesam_eltype_2_general
from ada.fem.formats.utils import str_to_int
from ada.fem.shapes.lines import SpringTypes

from . import cards

logger = get_logger()


def get_elements(bulk_str: str, fem: FEM) -> tuple[FemElements, dict, dict, dict]:
    """Import elements from Sesam Bulk str"""

    mass_elem = dict()
    spring_elem = dict()
    internal_external_element_map = dict()

    def grab_elements(match):
        d = match.groupdict()
        el_no = str_to_int(d["elno"])
        el_nox = str_to_int(d["elnox"])
        internal_external_element_map[el_no] = el_nox
        nodes = [
            fem.nodes.from_id(x)
            for x in filter(
                lambda x: x != 0,
                map(str_to_int, d["nids"].replace("\n", "").split()),
            )
        ]
        eltyp = d["eltyp"]
        el_type = sesam_eltype_2_general(str_to_int(eltyp))

        if isinstance(el_type, SpringTypes):
            spring_elem[el_no] = dict(gelmnt=d)
            return None

        metadata = dict(eltyad=str_to_int(d["eltyad"]), eltyp=eltyp)
        elem = Elem(el_no, nodes, el_type, None, parent=fem, metadata=metadata)

        if el_type == Elem.EL_TYPES.MASS_SHAPES.MASS:
            logger.warning("Mass element interpretation in sesam is undergoing changes. Results should be checked")
            mass_elem[el_no] = dict(gelmnt=d)
            fem.sets.add(FemSet(f"m{el_no}", [elem], FemSet.TYPES.ELSET, parent=fem))

        return elem

    elements = FemElements(
        filter(lambda x: x is not None, map(grab_elements, cards.GELMNT1.to_ff_re().finditer(bulk_str))), fem_obj=fem
    )
    return elements, mass_elem, spring_elem, internal_external_element_map


def get_mass(bulk_str: str, fem: FEM, mass_elem: dict) -> FemElements:
    def checkEqual2(iterator):
        return len(set(iterator)) <= 1

    def find_bnmass(match) -> Mass:
        d = match.groupdict()

        nodeno = str_to_int(d["nodeno"])
        mass_in = [
            roundoff(d["m1"]),
            roundoff(d["m2"]),
            roundoff(d["m3"]),
            roundoff(d["m4"]),
            roundoff(d["m5"]),
            roundoff(d["m6"]),
        ]
        masses = [m for m in mass_in if m != 0.0]
        if checkEqual2(masses):
            mass_type = Mass.PTYPES.ISOTROPIC
            masses = [masses[0]] if len(masses) > 0 else [0.0]
        else:
            mass_type = Mass.PTYPES.ANISOTROPIC

        no = fem.nodes.from_id(nodeno)
        fem_set = fem.sets.add(FemSet(f"m{nodeno}", [no], FemSet.TYPES.NSET, parent=fem))
        el_id = fem.elements.max_el_id + 1
        elem = fem.elements.add(Elem(el_id, [no], Elem.EL_TYPES.MASS_SHAPES.MASS, None, parent=fem), skip_grouping=True)
        mass = Mass(f"m{nodeno}", fem_set, masses, Mass.TYPES.MASS, ptype=mass_type, parent=fem, mass_id=el_id)

        elset = fem.sets.add(FemSet(f"m{nodeno}", [elem], FemSet.TYPES.ELSET, parent=fem))
        elem.mass_props = mass
        elem.elset = elset
        return mass

    def find_mgmass(match) -> Mass:
        d = match.groupdict()
        matno = str_to_int(d["matno"])
        mat_mass_map = {str_to_int(val["section_data"]["matno"]): val for key, val in mass_elem.items()}
        mass_el: dict = mat_mass_map.get(matno, None)
        if mass_el is None:
            raise ValueError()
        ndof = str_to_int(d["ndof"])
        if ndof != 6:
            raise NotImplementedError("Only mass matrices with 6 DOF are currently supported for reading")

        r = [float(x) for x in d["bulk"].split()]
        A = np.matrix(
            [
                [r[0], 0.0, 0.0, 0.0, 0.0, 0.0],
                [r[1], r[6], 0.0, 0.0, 0.0, 0.0],
                [r[2], r[7], r[11], 0.0, 0.0, 0.0],
                [r[3], r[8], r[12], r[15], 0.0, 0.0],
                [r[4], r[9], r[13], r[16], r[18], 0.0],
                [r[5], r[10], r[14], r[17], r[19], r[20]],
            ]
        )
        # use symmetry to complete the 6x6 matrix
        mass_matrix_6x6 = np.tril(A) + np.triu(A.T, 1)
        nodeno = str_to_int(mass_el["gelmnt"].get("nids"))
        elno = str_to_int(mass_el["gelmnt"].get("elno"))
        no = fem.nodes.from_id(nodeno)
        fem_set = fem.sets.add(FemSet(f"m{nodeno}", [no], FemSet.TYPES.NSET, parent=fem))

        mass_type = Mass.PTYPES.ANISOTROPIC
        mass = Mass(f"m{nodeno}", fem_set, mass_matrix_6x6, Mass.TYPES.MASS, ptype=mass_type, parent=fem, mass_id=elno)
        mass_el["el"] = mass
        return mass

    bn_masses = map(find_bnmass, cards.re_bnmass.finditer(bulk_str))
    mg_masses = map(find_mgmass, cards.re_mgmass.finditer(bulk_str))
    return FemElements(chain(bn_masses, mg_masses), fem_obj=fem)


def get_springs(bulk_str, fem: FEM, spring_elem: dict):
    matno_map = {str_to_int(sp["section_data"]["matno"]): sp for sp in spring_elem.values()}

    def find_mgspring(m):
        nonlocal matno_map
        d = m.groupdict()
        matno = str_to_int(d["matno"])
        ndof = str_to_int(d["ndof"])
        res: dict = matno_map.get(matno, None)
        if res is None:
            raise ValueError()

        elid = str_to_int(res["section_data"]["elno"])
        bulk = d["bulk"].replace("\n", "").split()

        spr_name = f"spr{elid}"
        nid = res["gelmnt"].get("nids", None)
        n1 = fem.nodes.from_id(str_to_int(nid))
        a = 1
        row = 0
        spring = []
        subspring = []
        for dof in bulk:
            subspring.append(float(dof.strip()))
            a += 1
            if a > ndof - row:
                spring.append(subspring)
                subspring = []
                a = 1
                row += 1
        new_s = []
        for row in spring:
            l = abs(len(row) - 6)
            if l > 0:
                new_s.append([0.0 for i in range(0, l)] + row)
            else:
                new_s.append(row)
        spring_matrix = np.array(new_s)
        spring_matrix = spring_matrix + spring_matrix.T - np.diag(np.diag(spring_matrix))
        fs = FemSet(f"{spr_name}_set", [n1], FemSet.TYPES.NSET, parent=fem)
        return Spring(spr_name, elid, "SPRING1", fem_set=fs, stiff=spring_matrix, parent=fem)

    return {c.name: c for c in map(find_mgspring, cards.re_mgsprng.finditer(bulk_str))}



import xml.etree.ElementTree as ET

from ada import Node, Part
from ada.fem import FemSet, Mass


def get_masses(xml_root: ET.Element, parent: Part) -> dict[str, Mass]:
    masses = dict()
    for sp in xml_root.findall(".//point_mass"):
        name = sp.attrib.get("name")

        # Get position
        pos = sp.findall(".//position")[0]
        n = Node([float(y) for x, y in pos.items()])
        parent.fem.nodes.add(n)

        # Get mass
        mass_res = sp.findall(".//mass_scalar")[0]
        mass_value = float(mass_res.attrib.get("mass"))
        fs = FemSet(f"{name}_fs", [n])
        masses[name] = Mass(name, ref=fs, mass=float(mass_value))

    return masses



# coding=utf-8
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np

from ada.base.units import Units

if TYPE_CHECKING:
    from ada import Material

from .plasticity_models import PlasticityModel


@dataclass
class MaterialDampingRayleigh:
    alpha: float = field(default=None, init=False)
    beta: float = field(default=None, init=False)


class Metal:
    """Base object for all metals"""

    def __init__(self, E, rho, sig_y, sig_u, v, zeta, alpha, plasticitymodel=PlasticityModel(), units="m", parent=None):
        self._E = E
        self._rho = rho
        self._sig_y = sig_y
        self._sig_u = sig_u
        self._v = v
        self.zeta = zeta
        self._alpha = alpha
        self._plasticity_model = plasticitymodel
        self._units = units
        self._rayleigh_damping = MaterialDampingRayleigh()
        self._parent = parent

    def __delattr__(self, item):
        raise AttributeError("Deletion of base material object properties is not allowed!")

    def __eq__(self, other: Metal):
        for att in filter(lambda x: x.startswith("__") is False and not callable(getattr(self, x)), dir(self)):
            self_var = getattr(self, att)
            other_var = getattr(other, att)
            if att in ["GRADES", "EC3_E_RED", "EC3_S_RED", "EC3_TEMP"]:
                continue
            if type(self_var) in (float, str, int) or self_var is None:
                return self_var == other_var
            elif type(self_var) in (list,):
                return all([x == y for x, y in zip(self_var, other_var)])
            elif type(self_var) in (np.ndarray,):
                comparison = self_var == other_var
                return comparison.all()
            else:
                raise NotImplementedError()

        return True

    def equal_props(self, other: Metal):
        for pa, pb in zip(self.unique_props(), other.unique_props()):
            if pa != pb:
                return False

        return True

    def unique_props(self):
        props = [
            "E",
            "sig_y",
            "sig_u",
            "rho",
            "v",
            "alpha",
            "zeta",
            "plasticity_model",
            "E_therm",
            "sigy_therm",
            "kappa",
            "rayleigh_damping",
            "cp",
        ]
        return [getattr(self, p) for p in props]

    def __repr__(self):
        return f"Metal(E:{self.E}, rho:{self.rho}, Sigy: {self.sig_y}, Plasticity Model: {self.plasticity_model})"

    @property
    def E(self):
        """Young's Modulus"""
        return self._E

    @E.setter
    def E(self, value):
        if value < 0:
            raise ValueError("The Young's Modulus must be a positive number")
        self._E = value

    @property
    def G(self):
        """Shear Modulus"""
        return self.E / (2 * (1 + self.v))

    @property
    def sig_y(self):
        """Yield stress"""
        return self._sig_y

    @sig_y.setter
    def sig_y(self, value):
        if value < 0:
            raise ValueError("Yield Stress must be a positive number")
        self._sig_y = value

    @property
    def sig_u(self):
        """Ultimate yield stress"""
        return self._sig_u

    @property
    def rho(self) -> float:
        """Density"""
        return self._rho

    @rho.setter
    def rho(self, value: float):
        if value < 0.0:
            raise ValueError("Material density must be a positive number")
        self._rho = value

    @property
    def v(self) -> float:
        """Poisson Ratio"""
        return self._v

    @property
    def alpha(self) -> float:
        """Thermal Expansion coefficient"""
        return self._alpha

    @property
    def zeta(self) -> float:
        """Material damping coefficient"""
        return self._zeta

    @zeta.setter
    def zeta(self, value: float):
        if value is None or value < 0.0:
            raise ValueError("Zeta cannot be None or below zero")
        self._zeta = value

    @property
    def plasticity_model(self) -> PlasticityModel:
        """Constitutive Equation for plasticity"""
        return self._plasticity_model

    @plasticity_model.setter
    def plasticity_model(self, value: PlasticityModel):
        self._plasticity_model = value

    @property
    def E_therm(self):
        return None

    @property
    def sigy_therm(self):
        return None

    @property
    def kappa(self):
        return None

    @property
    def rayleigh_damping(self):
        return self._rayleigh_damping

    @property
    def cp(self):
        return None

    @property
    def units(self):
        return self._units

    @units.setter
    def units(self, value):
        if isinstance(value, str):
            value = Units.from_str(value)
        if self._units == Units.M and value == Units.MM:
            self._E *= 1e-6
            if self._sig_y is not None:
                self._sig_y *= 1e-6
            self._rho *= 1e-9
            self._units = value
        elif self._units == Units.MM and value == Units.M:
            self._E *= 1e6
            if self._sig_y is not None:
                self._sig_y *= 1e6
            self._rho *= 1e9
            self._units = value

    @property
    def parent(self) -> "Material":
        return self._parent

    @parent.setter
    def parent(self, value):
        self._parent = value


class CarbonSteelGradeTypes:
    S355 = "S355"
    S420 = "S420"


class CarbonSteel(Metal):
    TYPES = CarbonSteelGradeTypes
    GRADES = {
        TYPES.S355: dict(name="S355", sigy=355e6, sigu=355e6),
        TYPES.S420: dict(name="S420", sigy=420e6, sigu=420e6),
    }
    EC3_TEMP = [20, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 1100, 1200]
    EC3_E_RED = [
        1.0,
        1.0,
        0.9,
        0.8,
        0.7,
        0.6,
        0.31,
        0.13,
        0.09,
        0.0675,
        0.045,
        0.0225,
        0.0,
    ]
    EC3_S_RED = [1.0, 1.0, 1.0, 1.0, 1.0, 0.78, 0.47, 0.23, 0.11, 0.06, 0.04, 0.02, 0.0]

    def __init__(
        self,
        grade="S355",
        plasticity_model: PlasticityModel = None,
        E=2.1e11,
        rho=7850,
        v=0.3,
        zeta=1.15,
        alpha=1.2e-5,
        sig_y=None,
        sig_u=None,
        temp_range=None,
        units="m",
        parent=None,
    ):
        """
        :param grade: Material Grade
        :param plasticity_model: Plasticity model e.g. CarbonSteel
        :param E: Young's Modulus
        :param rho: Material Density
        :param v: Poisson Ratio
        :param zeta: Material damping coefficient
        :param alpha: Thermal Expansion coefficient
        :param sig_y: Yield stress
        :param sig_u: Ultimate stress
        :param temp_range: Temperature range
        :param units: Definition of length unit. Default is meter 'm'. Alternative is millimeter 'mm'.
        """
        self._grade = grade
        sig_y = sig_y if sig_y is not None else CarbonSteel.GRADES[grade]["sigy"]
        sig_u = sig_u if sig_u is not None else CarbonSteel.GRADES[grade]["sigu"]
        super(CarbonSteel, self).__init__(
            E=E,
            rho=rho,
            sig_y=sig_y,
            sig_u=sig_u,
            v=v,
            zeta=zeta,
            alpha=alpha,
            plasticitymodel=plasticity_model,
            units=units,
            parent=parent,
        )
        # Manually override variables

        self._temp_range = np.arange(20, 1210, 5) if temp_range is None else temp_range

    def __repr__(self):
        return (
            f"CarbonSteel(E:{self.E:.3E}, sig_y:{self.sig_y:.3E}, rho:{self.rho:.3E},"
            f" plasticity_model:{self.plasticity_model})"
        )

    @property
    def grade(self):
        """Material Grade"""
        return self._grade

    @property
    def temp_range(self):
        return self._temp_range

    @property
    def E_therm(self):
        E_red_fac = np.interp(self._temp_range, CarbonSteel.EC3_TEMP, CarbonSteel.EC3_E_RED)
        return [self._E * x for x in E_red_fac]

    @property
    def sigy_therm(self):
        sig_red_fac = np.interp(self._temp_range, CarbonSteel.EC3_TEMP, CarbonSteel.EC3_S_RED)
        return [self._sig_y * x for x in sig_red_fac]

    @property
    def kappa(self):
        """Thermal conductivity. Watts per meter-kelvin W/(mK)"""
        phase1_end = 780
        phase1_arr = [self._temp_range[x] for x in np.where(self._temp_range <= phase1_end)]
        phase2_arr = [self._temp_range[x] for x in np.where(self._temp_range > phase1_end)]
        phase1 = [54 - t * 3.33 * 0.01 for t in phase1_arr[0]]
        phase2 = [27.3 for x in range(phase2_arr[0].shape[0])]
        return phase1 + phase2

    @property
    def cp(self):
        """Specific Heat. Joule per kelvin and kilogram J/(K kg)"""

        phase1_end = 600
        phase2_end = 735
        phase3_end = 900
        phase1_arr = [self._temp_range[x] for x in np.where(self._temp_range <= phase1_end)]
        phase2_arr = [
            self._temp_range[x]
            for x in np.where(np.logical_and(self._temp_range > phase1_end, self._temp_range <= phase2_end))
        ]
        phase3_arr = [
            self._temp_range[x]
            for x in np.where(np.logical_and(self._temp_range > phase2_end, self._temp_range <= phase3_end))
        ]
        phase4_arr = [self._temp_range[x] for x in np.where(self._temp_range > phase3_end)]
        phase1 = [425 + 7.73 * 1e-1 * t - 1.69 * 1e-3 * (t**2) + 2.22 * 1e-6 * t**3 for t in phase1_arr[0]]
        phase2 = [666 + 13002 / (738 - t) for t in phase2_arr[0]]
        phase3 = [545 + 17820 / (t - 731) for t in phase3_arr[0]]
        phase4 = [650 for x in range(phase4_arr[0].shape[0])]
        return phase1 + phase2 + phase3 + phase4


class Aluminium(Metal):
    def __init__(self):
        raise NotImplementedError("The aluminium material model is not yet implemented")



from itertools import groupby
from operator import attrgetter
from typing import Iterable

from ada.core.utils import NewLine
from ada.fem import Elem, FemSection
from ada.fem.containers import FemElements
from ada.fem.shapes import ElemShape
from ada.fem.shapes import definitions as shape_def


def elements_str(fem_elements: FemElements) -> str:
    if len(fem_elements) == 0:
        return "** No elements"

    el_str = ""
    for (el_type, fem_sec), elements in groupby(fem_elements, key=attrgetter("type", "fem_sec")):
        if isinstance(el_type, shape_def.ConnectorTypes):
            continue
        el_str += elwriter(el_type, fem_sec, elements)

    return el_str


def elwriter(eltype, fem_sec: FemSection, elements: Iterable[Elem]):
    sub_eltype = el_type_sub(eltype, fem_sec)
    el_set_str = f", ELSET={fem_sec.elset.name}" if fem_sec.elset is not None else ""
    el_str = "\n".join((write_elem(el) for el in elements))

    return f"""*ELEMENT, type={sub_eltype}{el_set_str}\n{el_str}\n"""


def el_type_sub(el_type, fem_sec: FemSection) -> str:
    """Substitute Element types specifically Calculix"""

    if isinstance(el_type, shape_def.LineShapes):
        if must_be_converted_to_general_section(fem_sec.section.type):
            return "U1"
    if el_type == ElemShape.TYPES.shell.TRI6:
        return "S6"
    fem = fem_sec.parent
    return fem.options.ABAQUS.default_elements.get_element_type(el_type)


def must_be_converted_to_general_section(sec_type):
    from ada.sections.categories import BaseTypes

    if sec_type in [BaseTypes.CIRCULAR, BaseTypes.IPROFILE, BaseTypes.GENERAL, BaseTypes.ANGULAR]:
        return True
    else:
        return False


def write_elem(el: Elem) -> str:
    nl = NewLine(10, suffix=7 * " ")
    if len(el.nodes) > 6:
        di = " {}"
    else:
        di = "{:>13}"
    el_str = f"{el.id:>7}, " + " ".join([f"{di.format(no.id)}," + next(nl) for no in el.nodes])[:-1]
    return el_str



from __future__ import annotations

import os
from typing import TYPE_CHECKING, List

from ada.fem.formats.utils import DatFormatReader

if TYPE_CHECKING:
    from ada.fem.results.eigenvalue import EigenDataSummary


def get_eigen_data(dat_file: str | os.PathLike) -> EigenDataSummary:
    from ada.fem.results.eigenvalue import EigenDataSummary, EigenMode

    dtr = DatFormatReader()

    re_compiled = dtr.compile_ff_re([int] + [float] * 4)
    re_compiled_2 = dtr.compile_ff_re([int] + [float] * 6)
    re_compiled_3 = dtr.compile_ff_re([float] * 6)

    eig_str = "eigenvalueoutput"
    part_str = "participationfactors"
    eff_modal = "effectivemodalmass"
    tot_eff = "totaleffectivemass"

    eig_res = dtr.read_data_lines(dat_file, re_compiled, eig_str, part_str, split_data=True)
    part_res = dtr.read_data_lines(dat_file, re_compiled_2, part_str, eff_modal, split_data=True)
    modalmass = dtr.read_data_lines(dat_file, re_compiled_2, eff_modal, tot_eff, split_data=True)
    tot_eff_mass = dtr.read_data_lines(dat_file, re_compiled_3, tot_eff, split_data=True)[0]

    dof_base = ["x", "y", "z", "rx", "ry", "rz"]
    part_factor_names = ["p" + x for x in dof_base]
    eff_mass_names = ["ef" + x for x in dof_base]

    eigen_modes: List[EigenMode] = []
    # Note! participation factors and effective modal mass are each deconstructed into 6 degrees of freedom
    for eig, part, modal in zip(eig_res, part_res, modalmass):
        mode, eig_value, freq_rad, freq_cycl, freq_imag_rad = eig
        eig_output = dict(
            eigenvalue=float(eig_value),
            f_rad=float(freq_rad),
            f_hz=float(freq_cycl),
            f_imag_rad=float(freq_imag_rad),
        )
        participation_data = {pn: p for pn, p in zip(part_factor_names, part[1:])}
        eff_mass_data = {pn: p for pn, p in zip(eff_mass_names, part[1:])}
        eigen_modes.append(EigenMode(no=mode, **eig_output, **participation_data, **eff_mass_data))

    return EigenDataSummary(eigen_modes, tot_eff_mass)



from __future__ import annotations

import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

from ada.config import get_logger

if TYPE_CHECKING:
    from ada import Part, Section

logger = get_logger()


def add_sections(root: ET.Element, part: Part):
    sections_elem = ET.Element("sections")

    # Add the new element underneath <properties>
    root.append(sections_elem)

    for section in part.sections:
        if section.type == section.TYPES.ANGULAR:
            add_angular_section(section, sections_elem)
        elif section.type == section.TYPES.TUBULAR:
            add_pipe_section(section, sections_elem)
        elif section.type == section.TYPES.IPROFILE and section.w_btn == section.w_top:
            add_i_section(section, sections_elem)
        elif section.type == section.TYPES.IPROFILE and section.w_btn != section.w_top:
            add_unsymm_i_section(section, sections_elem)
        elif section.type == section.TYPES.BOX:
            add_box_section(section, sections_elem)
        else:
            logger.error(f"The profile type {section.type} is not yet supported for Genie XML export")


def add_angular_section(section: Section, xml_root: ET.Element):
    section_elem = ET.Element("section", {"name": section.name, "description": ""})
    section_props = ET.Element(
        "l_section",
        dict(
            h=str(section.h),
            b=str(section.w_btn),
            tw=str(section.t_w),
            tf=str(section.t_fbtn),
            fabrication="unknown",
            sfy="1",
            sfz="1",
            general_properties_method="computed",
        ),
    )

    section_elem.append(section_props)
    xml_root.append(section_elem)


def add_pipe_section(section: Section, xml_root: ET.Element):
    section_elem = ET.Element("section", {"name": section.name, "description": ""})

    section_props = ET.Element(
        "pipe_section",
        dict(
            od=str(section.r * 2),
            th=str(section.wt),
            fabrication="unknown",
            sfy="1",
            sfz="1",
            general_properties_method="computed",
        ),
    )

    section_elem.append(section_props)
    xml_root.append(section_elem)


def add_i_section(section: Section, xml_root: ET.Element):
    section_elem = ET.Element("section", {"name": section.name, "description": ""})
    section_props = ET.Element(
        "i_section",
        dict(
            h=str(section.h),
            b=str(section.w_top),
            tw=str(section.t_w),
            tf=str(section.t_fbtn),
            fillet_radius="0.024",
            fabrication="unknown",
            sfy="1",
            sfz="1",
            general_properties_method="computed",
        ),
    )

    section_elem.append(section_props)
    xml_root.append(section_elem)


def add_box_section(section: Section, xml_root: ET.Element):
    # Create the <section> element
    xml_section = ET.Element("section", {"name": section.name})
    box_section = ET.Element(
        "box_section",
        dict(
            h=str(section.h),
            b=str(section.w_btn),
            tw=str(section.t_w),
            tfbot=str(section.t_fbtn),
            tftop=str(section.t_ftop),
            fabrication="unknown",
            sfy="1",
            sfz="1",
            general_properties_method="computed",
        ),
    )

    # Append the <box_section> element to the <section> element
    xml_section.append(box_section)
    xml_root.append(xml_section)
    # You can now append the <section> element to your xml tree


def add_unsymm_i_section(section: Section, xml_root: ET.Element):
    section_elem = ET.Element("section", {"name": section.name})
    section_props = ET.Element(
        "unsymmetrical_i_section",
        dict(
            h=str(section.h),
            tw=str(section.t_w),
            bftop=str(section.w_top),
            bftop1=str(section.w_top / 2),
            bfbot=str(section.w_btn),
            bfbot1=str(section.w_btn / 2),
            tfbot=str(section.t_fbtn),
            tftop=str(section.t_ftop),
            fabrication="unknown",
            sfy="1",
            sfz="1",
            general_properties_method="computed",
        ),
    )

    section_elem.append(section_props)
    xml_root.append(section_elem)



from __future__ import annotations

import xml.etree.ElementTree as ET
from typing import TYPE_CHECKING

from ada.core.constants import X, Y, Z

from .write_utils import add_local_system

if TYPE_CHECKING:
    from ada import Part


def add_masses(root: ET.Element, part: Part):
    all_mass = list(part.fem.get_all_masses())
    for mass in all_mass:
        if len(mass.fem_set.members) != 1:
            raise NotImplementedError()

        n = mass.fem_set.members[0]

        bc_stru = ET.SubElement(root, "structure")
        sup_point = ET.SubElement(bc_stru, "point_mass", {"name": mass.name})
        sup_point.append(add_local_system(X, Y, Z))
        geom = ET.SubElement(sup_point, "geometry")
        ET.SubElement(geom, "position", {"x": str(n.x), "y": str(n.y), "z": str(n.z)})
        bc_con = ET.SubElement(sup_point, "mass")
        ET.SubElement(bc_con, "mass_scalar", dict(mass=str(mass.mass)))



from __future__ import annotations

import pathlib
from dataclasses import dataclass
from typing import TYPE_CHECKING

import h5py
import numpy as np

from ada.config import get_logger

if TYPE_CHECKING:
    from ada.fem.results.common import (
        ElementBlock,
        ElementFieldData,
        FEAResult,
        FemNodes,
        Mesh,
        NodalFieldData,
    )

logger = get_logger()


def read_rmed_file(rmed_file: str | pathlib.Path) -> FEAResult:
    from ada.fem.results.common import FEAResult, FEATypes

    if isinstance(rmed_file, str):
        rmed_file = pathlib.Path(rmed_file)

    mr = MedReader(rmed_file)
    mr.load()

    return FEAResult(
        rmed_file.name, software=FEATypes.CODE_ASTER, results=mr.results, mesh=mr.mesh, results_file_path=rmed_file
    )


@dataclass
class MedReader:
    rmed_file: str | pathlib.Path
    f: h5py.File = None

    mesh: Mesh = None
    results: list[ElementFieldData | NodalFieldData] = None

    _dim: int = None
    is_eigen_analysis: bool = False

    def __post_init__(self):
        if isinstance(self.rmed_file, str):
            self.rmed_file = pathlib.Path(self.rmed_file)

    def load(self):
        self.f = h5py.File(self.rmed_file)
        self.mesh = self.get_mesh()
        self.results = self.get_results()
        self.f.close()

    def get_mesh(self) -> Mesh:
        from ada.fem.results.common import Mesh

        mesh = self._load_mesh()

        nodes = self.get_nodes(mesh)
        elements = self.get_elements(mesh)

        return Mesh(elements=elements, nodes=nodes)

    def get_nodes(self, mesh) -> FemNodes:
        from ada.fem.results.common import FemNodes

        pts_dataset = mesh["NOE"]["COO"]
        n_points = pts_dataset.attrs["NBR"]
        coords = pts_dataset[()].reshape((n_points, self._dim), order="F")
        node_identifiers = np.array(range(1, len(coords) + 1))

        # TODO: Figure out where the actual point identifiers are stored in med files
        return FemNodes(coords=coords, identifiers=node_identifiers)

    def get_elements(self, mesh) -> list[ElementBlock]:
        from ada.fem.formats.code_aster.common import med_to_ada_type
        from ada.fem.results.common import ElementBlock, ElementInfo, FEATypes

        cell_types = []
        med_cells = mesh["MAI"]

        blocks = []
        for med_cell_type, med_cell_type_group in med_cells.items():
            if med_cell_type == "PO1":
                logger.warning("Point elements are still not supported")
                continue

            cell_type = med_to_ada_type(med_cell_type)
            cell_types.append(cell_type)

            nod = med_cell_type_group["NOD"]
            n_cells = nod.attrs["NBR"]
            node_refs = nod[()].reshape(n_cells, -1, order="F")

            if "NUM" in med_cell_type_group.keys():
                num = np.array(med_cell_type_group["NUM"])
            else:
                num = np.arange(0, len(node_refs))

            elem_info = ElementInfo(type=cell_type, source_software=FEATypes.CODE_ASTER, source_type=med_cell_type)
            elem_block = ElementBlock(elem_info=elem_info, node_refs=node_refs, identifiers=num)
            blocks.append(elem_block)

        return blocks

    def get_results(self) -> list[ElementFieldData | NodalFieldData]:
        fields = self.f.get("CHA")
        results = []

        for name, data in fields.items():
            nom = data.attrs.get("NOM")
            if nom is None:
                raise ValueError()

            components = nom.decode().split()
            time_step = sorted(data.keys())  # associated time-steps
            time_steps = []
            if len(time_step) == 1:  # single time-step
                names = [name]  # do not change field name
                res = data[time_step[0]].attrs.get("PDT")
                time_steps.append(res)
            else:  # many time-steps
                names = [None] * len(time_step)
                for i, key in enumerate(time_step):
                    t = data[key].attrs["PDT"]  # current time
                    time_steps.append(float(t))
                    names[i] = name + f"[{i:d}] - {t:g}"

            if name == "modes___DEPL":
                self.is_eigen_analysis = True

            # MED field can contain multiple types of data
            for i, key in enumerate(time_step):
                med_data = data[key]  # at a particular time step
                step_name = names[i]
                ts = time_steps[i]
                step_index = i + 1
                for supp in med_data:
                    if supp == "NOE":  # continuous nodal (NOEU) data
                        result = self._load_nodal_field_data(step_index, step_name, med_data, ts, components)
                        results.append(result)
                    else:  # Gauss points (ELGA) or DG (ELNO) data
                        result = self._load_element_field_data(step_index, step_name, med_data[supp], ts, components)
                        results.append(result)

        return results

    def _load_element_field_data(self, i, name, med_data, step, components) -> ElementFieldData:
        from ada.fem.results.common import ElementFieldData

        profile = med_data.attrs["PFL"]
        data_profile = med_data[profile]
        n_cells = data_profile.attrs["NBR"]
        n_gauss_points = data_profile.attrs["NGA"]
        if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
            values = data_profile["CO"][()].reshape(n_cells, n_gauss_points, -1, order="F")
        else:
            raise NotImplementedError()

        # Only 1 data point per cell, shape -> (n_cells, n_components)
        if n_gauss_points == 1:
            values = values[:, 0, :]
            if values.shape[-1] == 1:  # cut off for scalars
                values = values[:, 0]

        eig_freq = None
        if self.is_eigen_analysis:
            eig_freq = step
            step = i

        return ElementFieldData(name, step, components, values, eigen_freq=eig_freq)

    def _load_nodal_field_data(self, i, name, med_data, step, components) -> NodalFieldData:
        from ada.fem.results.common import NodalFieldData, NodalFieldType

        profile = med_data["NOE"].attrs["PFL"]
        data_profile = med_data["NOE"][profile]
        n_points = data_profile.attrs["NBR"]
        if profile.decode() == "MED_NO_PROFILE_INTERNAL":  # default profile with everything
            values = data_profile["CO"][()].reshape(n_points, -1, order="F")
        else:
            raise NotImplementedError()

        if values.shape[-1] == 1:  # cut off for scalars
            values = values[:, 0]

        node_ids = self.mesh.nodes.identifiers
        values = np.insert(values, 0, node_ids, axis=1)

        eig_freq = None
        if self.is_eigen_analysis:
            eig_freq = step
            step = i

        field_type = None
        if "DX" in components:
            field_type = NodalFieldType.DISP

        return NodalFieldData(name, step, components, values, eigen_freq=eig_freq, field_type=field_type)

    def _load_mesh(self):
        mesh_ensemble = self.f["ENS_MAA"]
        meshes = mesh_ensemble.keys()
        if len(meshes) != 1:
            raise ValueError("Must only contain exactly 1 mesh, found {}.".format(len(meshes)))

        mesh_name = list(meshes)[0]
        mesh = mesh_ensemble[mesh_name]
        self._dim = mesh.attrs["ESP"]

        # Possible time-stepping
        if "NOE" not in mesh:
            # One needs NOE (node) and MAI (French maillage, meshing) data. If they
            # are not available in the mesh, check for time-steppings.
            time_step = mesh.keys()
            if len(time_step) != 1:
                raise ValueError(f"Must only contain exactly 1 time-step, found {len(time_step)}.")
            mesh = mesh[list(time_step)[0]]
        return mesh



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, List, Union

import numpy as np


@dataclass
class DataFilter:
    name_filter: Union[None, List[str]] = None
    filter_elements_by_guid: Union[None, List[str]] = None
    filter_func: Callable = None
    filter_func_ref: str = None


@dataclass
class ExportConfig:
    quality: float = 1.0
    threads: int = 1
    parallel: bool = True
    merge_by_colour: bool = False
    merge_subgeometries_by_colour: bool = True
    render_edges: bool = False
    ifc_skip_occ: bool = True
    data_filter: DataFilter = field(default_factory=DataFilter)

    # Position of model
    volume_center: Union[None, np.ndarray] = None
    auto_center_model: bool = False
    max_convert_objects: Union[int, None] = None
    do_not_load_by_default: List[str] = None
    use_cache: bool = False

    metadata: dict = field(default_factory=dict)
    name_prefix: str = ""



from __future__ import annotations

import pathlib
from io import StringIO
from typing import TYPE_CHECKING, Tuple, Union

import ifcopenshell
import ifcopenshell.geom
from ifcopenshell.util.element import get_psets

from ada.concepts.transforms import Placement
from ada.config import Settings, get_logger

if TYPE_CHECKING:
    from ada import Assembly, Part, Pipe

tol_map = dict(m=Settings.mtol, mm=Settings.mmtol)
logger = get_logger()


def open_ifc(ifc_file_path: Union[str, pathlib.Path, StringIO]):
    if type(ifc_file_path) is StringIO:
        return ifcopenshell.file.from_string(str(ifc_file_path.read()))
    return ifcopenshell.open(str(ifc_file_path))


def get_ifc_property_sets(ifc_elem) -> dict:
    """Returns a dictionary of {pset_id:[prop_id, prop_id...]} for an IFC object"""
    props = dict()
    for definition in ifc_elem.IsDefinedBy:
        if definition.is_a("IfcRelDefinesByProperties") is False:
            continue
        property_set = definition.RelatingPropertyDefinition
        if property_set.is_a("IfcElementQuantity"):
            continue

        pset_name = property_set.Name.split(":")[0].strip()
        props[pset_name] = dict()
        for prop in property_set.HasProperties:
            if prop.is_a() not in ("IfcPropertySingleValue", "IfcPropertyListValue"):
                continue
            if prop.is_a("IfcPropertySingleValue"):
                res = prop.NominalValue.wrappedValue if prop.NominalValue is not None else None
                props[pset_name][prop.Name] = res
            else:
                props[pset_name][prop.Name] = [x.wrappedValue for x in prop.ListValues]

    return props


def get_parent(instance):
    from ifcopenshell.util.element import get_container

    if instance.is_a("IfcOpeningElement"):
        return instance.VoidsElements[0].RelatingBuildingElement
    if instance.is_a("IfcElement"):
        fills = instance.FillsVoids
        if len(fills):
            return fills[0].RelatingOpeningElement
        containments = instance.ContainedInStructure
        if len(containments):
            return containments[0].RelatingStructure
    if instance.is_a("IfcObjectDefinition"):
        decompositions = instance.Decomposes
        if len(decompositions):
            return decompositions[0].RelatingObject

    return get_container(instance)


def get_associated_material(ifc_elem: ifcopenshell.entity_instance):
    c = None
    for association in ifc_elem.HasAssociations:
        if association.is_a("IfcRelAssociatesMaterial") is False:
            continue
        material = association.RelatingMaterial
        if material.is_a("IfcMaterialProfileSet"):
            # For now, we only deal with a single profile
            c = material.MaterialProfiles[0]
        if material.is_a("IfcMaterialProfileSetUsage"):
            c = material.ForProfileSet.MaterialProfiles[0]
        if material.is_a("IfcRelAssociatesMaterial"):
            c = material.RelatingMaterial
        if material.is_a("IfcMaterial"):
            c = material

    if c is None:
        raise ValueError(f"{ifc_elem=} lacks associated Material properties")

    return c


def get_beam_type(ifc_elem: ifcopenshell.entity_instance) -> ifcopenshell.entity_instance:
    for typed_by in ifc_elem.IsTypedBy:
        if typed_by.RelatingType.is_a("IfcBeamType"):
            return typed_by.RelatingType


def get_name_from_props(props: dict) -> str | None:
    name = None
    for key, val in props.items():
        if type(val) is dict:
            name = get_name_from_props(val)
            if name is not None:
                break
        else:
            if key.lower() == "name":
                name = val
                break
    return name


def resolve_name(props, product):
    if product.Name is not None:
        return product.Name

    if hasattr(product, "Tag"):
        if product.Tag is not None:
            return product.Tag

    # This procedure is just to handle reading badly created ifc files with little or no related names
    name = get_name_from_props(props)
    if name is not None:
        return name

    logger.debug(f'Name/tag not found for ifc element "{product}". Using GlobalID as name')
    return product.GlobalId


def get_person(f, user_id):
    for p in f.by_type("IfcPerson"):
        if p.Identification == user_id:
            return p
    return None


def get_org(f, org_id):
    for p in f.by_type("IfcOrganization"):
        if p.Identification == org_id:
            return p
    return None


def add_to_assembly(assembly: Assembly, obj, ifc_parent, elements2part):
    from ada import Pipe

    pp_name = ifc_parent.Name

    if pp_name is None:
        pp_name = resolve_name(get_psets(ifc_parent), ifc_parent)
        if pp_name is None:
            raise ValueError(f'Name of ifc element "{ifc_parent}" is None')

    imported = False
    if elements2part is not None:
        add_to_parent(assembly, obj)
        imported = True
    else:
        res = assembly.get_by_name(pp_name)
        if res is not None:
            add_to_parent(res, obj)
            imported = True

        if imported is False:
            for pipe in assembly.get_all_physical_objects(by_type=Pipe):
                if pipe.name == pp_name or pipe.metadata.get("original_name") == pp_name:
                    add_to_parent(pipe, obj)
                    imported = True
                    break

    if imported is False:
        logger.info(f'Unable to find parent "{pp_name}" for {type(obj)} "{obj.name}". Adding to Assembly')
        assembly.add_shape(obj)


def add_to_parent(parent: Part | Pipe, obj):
    from ada import Beam, Part, Pipe, PipeSegElbow, PipeSegStraight, Plate, Shape

    if type(obj) is Beam:
        parent.add_beam(obj)
    elif type(obj) is Plate:
        parent.add_plate(obj)
    elif issubclass(type(obj), Shape) and isinstance(parent, Part):
        parent.add_shape(obj)
    elif isinstance(obj, (PipeSegStraight, PipeSegElbow)) and isinstance(parent, Part):
        pipe = Pipe.from_segments(parent.name, [obj])
        parent.parent.add_pipe(pipe)
        parent.parent.parts.pop(parent.name)
    elif isinstance(obj, (PipeSegStraight, PipeSegElbow)) and isinstance(parent, Pipe):
        # Todo: PipeSegments should not really be resolved here.
        obj.parent = parent
        parent.segments.append(obj)
    elif isinstance(obj, Pipe):
        parent.add_pipe(obj)
    else:
        raise NotImplementedError("")


def get_point(cartesian_point) -> Tuple[float, float, float]:
    return cartesian_point.Coordinates


def get_direction(ifc_direction) -> Tuple[float, float, float]:
    return ifc_direction.DirectionRatios


def get_placement(ifc_position) -> Placement:
    origin = get_point(ifc_position.Location)
    xdir = get_direction(ifc_position.RefDirection)
    zdir = get_direction(ifc_position.Axis)

    return Placement(origin, xdir=xdir, zdir=zdir)


def get_axis_polyline_points_from_product(product) -> list[tuple[float, float, float]]:
    axis_data = []
    for axis in filter(lambda x: x.RepresentationIdentifier == "Axis", product.Representation.Representations):
        if len(axis.Items) != 1:
            raise ValueError("Axis should only contain 1 item")
        for cartesian_point in axis.Items[0].Points:
            axis_data.append(cartesian_point.Coordinates)

    return axis_data


def get_ifc_body(product, allow_multiple=False) -> ifcopenshell.entity_instance:
    bodies = []
    for body in filter(lambda x: x.RepresentationIdentifier != "Axis", product.Representation.Representations):
        if len(body.Items) != 1:
            raise ValueError("Axis should only contain 1 item")
        bodies.append(body.Items[0])

    if len(bodies) != 1:
        if allow_multiple is False:
            raise ValueError("Currently do not support multi body IFC products")
        else:
            return bodies

    return bodies[0]


def get_swept_area(product: ifcopenshell.entity_instance) -> ifcopenshell.entity_instance:
    from .exceptions import UnableToConvertBoolResToBeamException

    body = get_ifc_body(product)
    if body.is_a("IfcBooleanResult"):
        raise UnableToConvertBoolResToBeamException(f"Unable to convert {product} to beam")

    return body.SweptArea



import re

from .helper_utils import AbaFF

_re_in = re.IGNORECASE | re.MULTILINE | re.DOTALL

# Elements
re_el = re.compile(
    r"^\*Element,\s*type=(?P<eltype>.*?)(?:\n|,\s*elset=(?P<elset>.*?)\s*\n)(?<=)(?P<members>(?:.*?)(?=\*|\Z))",
    _re_in,
)

re_sets = re.compile(
    r"(?:\*(nset|elset),\s*(?:nset|elset))=(.*?)(?:,\s*(internal\s*)|(?:))"
    r"(?:,\s*instance=(.*?)|(?:))(?:,\s*(generate)|(?:))\s*\n(?<=)((?:.*?)(?=\*|\Z))",
    _re_in,
)
# Boundary Conditions
re_bcs = re.compile(
    r"(?:\*\*\s*Name:\s*(?P<name>.*?)\s*Type:\s*(?P<type>.*?)\n|)" r"\*Boundary\n(?<=)(?P<content>(?:.*?)(?=\*|\Z))",
    _re_in,
)

# Parts
parts_matches = re.compile(r"\*Part, name=(?P<name>.*?)\n(?P<bulk_str>.*?)\*End Part", _re_in)
part_names = re.compile(r"\*\*\s*PART INSTANCE:\s*(.*?)\n(.*)", _re_in)

# Instances
inst_matches = re.compile(
    r"\*Instance, name=(?P<inst_name>.*?), part=(?P<part_name>.*?)\n(?P<bulk_str>.*?)\*End Instance", _re_in
)

# Sections
_re_offset = r"(?:, offset=(?P<offset>.*?)|)"
_re_controls = r"(?:, controls=(?P<controls>.*?)|)"
re_shell = re.compile(
    r"\*\Shell Section, elset"
    rf"=(?P<elset>.*?)\s*, material=(?P<material>.*?){_re_offset}{_re_controls}\s*\n(?P<t>.*?),"
    rf"(?P<int_points>.*?)$",
    _re_in,
)

re_beam = re.compile(
    r"\*Beam Section,\s*elset=(?P<elset>.*?)\s*,\s*material=(?P<material>.*?)\s*,\s*"
    r"(?:temperature=(?P<temperature>.*?),|)\s*(?:section=|sect=)(?P<sec_type>.*?)\n"
    r"(?P<line1>.*?)\n(?P<line2>.*?)$",
    _re_in,
)

re_solid = re.compile(
    r"(?:\*\s*Section:\s*(.*?)\n|)\*\Solid Section,\s*elset=(.*?)\s*,\s*material=(.*?)\s*$",
    _re_in,
)

# Contact
contact_pairs = AbaFF(
    "Contact Pair",
    [
        (
            "interaction=",
            "small sliding==|",
            "type=|",
            "adjust=|",
            "mechanical constraint=|",
            "geometric correction=|",
            "cpset=|",
        ),
        ("surf1", "surf2"),
    ],
    nameprop=("Interaction", "name"),
)

contact_general = AbaFF(
    "Contact",
    args=[()],
    subflags=[
        ("Contact Inclusions", [(), ("surf1", "surf2")]),
        ("Contact Property Assignment", [(), ("vara", "varb", "interaction")]),
        ("Contact Formulation", [("type=",), ("csurf1", "csurf2", "csurf_type")]),
        ("Contact Initialization Assignment", [(), ("ssurf1", "ssurf2", "cinit")]),
        ("Surface Property Assignment", [("property=",), ("bulk>",)]),
    ],
    # nameprop=("Interaction", "name"),
)

# Connectors
connector_behaviour = AbaFF(
    "Connector Behavior",
    [("name=",)],
    [("Connector Elasticity", [("nonlinear|", "component=|"), ("bulk>",)])],
)
connector_section = AbaFF("Connector Section", [("elset=", "behavior="), ("contype",), ("csys",)])

# Constraints
sh2so_re = AbaFF("Shell to Solid Coupling", [("constraint name=", "influence distance=|"), ("surf1", "surf2")])
rigid_bodies = AbaFF("Rigid Body", [("ref node=", "elset=")])
coupling = AbaFF(
    "Coupling",
    [("constraint name=", "ref node=", "surface=", "orientation=|")],
    [("Kinematic", [(), ("bulk>",)])],
)
tie = AbaFF("Tie", [("name=", "adjust="), ("surf1", "surf2")])
# Other
surface_smoothing = AbaFF("Surface Smoothing", [("name=",), ("bulk>",)])
surface = AbaFF("Surface", [("type=", "name=", "internal|"), ("bulk>",)])
orientation = AbaFF(
    "Orientation",
    [
        ("name=", "definition=|", "local directions=|", "system=|"),
        ("ax", "ay", "az", "bx|", "by|", "bz|", "|cx", "|cy", "|cz"),
        ("v1|", "v2|"),
    ],
)
